
#!/usr/bin/python

import re
import sys
import socket
from optparse import OptionParser
from ipv4 import *
from PBunnyServices import *
import threading

"""
PortBunny Option-Parser
=======================

Parses the command-line.

The dictionary of options and the list
of hosts to scan can then be retrieved
using get_options_dict and
get_hosts_to_scan respectively.

"""

IP_EXPRESSION_REGEX = '^([\d]{1,3}\.){3}[\d]{1,3}(/[\d]{1,2})*$'
HOSTNAME_REGEX      = '^(([a-zA-Z0-9\-]+\.)+[a-z]+$)'

ip_re = re.compile(IP_EXPRESSION_REGEX)
hostname_re = re.compile(HOSTNAME_REGEX)


DEFAULT_TRIGGERS_TO_TEST = [['TCP_SYN', '80'], ['TCP_SYN', '22'],
                            ['TCP_SYN', '443'], ['TCP_SYN', '445'],
                            ['TCP_SYN', '113'], ['ICMP_ER', '0'],
                            ['ICMP_TS', '0'], ['UDP', '161'], ['UDP', '162']
                            ]

DEFAULT_NTRIGGER_JOBS_AT_ONCE = 40
DEFAULT_NFLOOD_JOBS_AT_ONCE = 1


class PBunnyOptionParser:

    ################### PUBLIC METHODS #########################


    """
    Create a PBunnyOptionParser by parsing
    the current command-line.
    
    """

    def __init__(self, services):        

        parser = OptionParser("")

        self.services = services
        
        parser.add_option("-p", "--ports", 
                          action = "callback", dest = "port_ranges",
                          default = self.get_default_ports_to_scan(),
                          callback = self.ports_option_parser,
                          help = "comma-seperated list of ports to scan")    

        parser.add_option('-d', '--discover', action="store_true",
                          dest = 'only_discover', default = False,
                          help = "Don't perform the actual scan. Only 'discover' hosts by sending triggers")
       
         
        
        parser.add_option("-t", '--triggers-to-test',
                          action = "callback", dest = "triggers_to_test",
                          default = DEFAULT_TRIGGERS_TO_TEST,
                          callback = self.triggers_to_test_parser,
                          help = "only the specified triggers will be tested. " +
                          "For example, '-t ICMP_ER 0 TCP_SYN 80' will only try an ICMP-ECHO-REQUEST " +
                          "and a TCP-SYN-packet on port 80"
                          )

        parser.add_option('-g', '--graph', action="store_true",
                          dest = 'create_graph', default = False,
                          help = "Create a graph of the congestion-window.")
       
                      
        (self.options, self.hosts) = parser.parse_args()            
        
        # self.hosts now contains the hosts to scan
        # validate these 'host-expressions' and
        # construct the final list of hosts.
        
        self.parse_host_expressions()
        
        # Initialize defaults for number of jobs at once
        self.options.ntrigger_jobs_at_once = DEFAULT_NTRIGGER_JOBS_AT_ONCE
        self.options.nflood_jobs_at_once = DEFAULT_NFLOOD_JOBS_AT_ONCE
       
        self.lock = threading.Semaphore(1)
        

    """
    returns the default-list of ports:
    well-known ports + all ports in the
    'services'-file
    """

    def get_default_ports_to_scan(self):
        retval = ['1-1024']

        ser = self.services.keys()
        ser.sort()

        for s in ser:
            if s > 1024:
                retval.append(str(s))
        
        return retval
    

    """
    Return list of hosts to scan:
        
    """

    def get_hosts_to_scan(self):
        return self.hosts
                            

    """
    Return dictionary of options.
    """
    
    def get_options_dict(self):
        return self.options


    """
    Generate a list of single ports
    from the given port-expressions
    and save it in self.options.ports_to_scan
    and its length in
    self.options.nports_to_scan_per_job
    
    """

    def gen_list_of_single_ports(self):
        retval = []

        for port_exp in self.options.port_ranges:

            # Since ports have been validated at construction,
            # this check is fine.
            
            match = re.compile('^([0-9]+)-([0-9]+)$').match(port_exp)
            
            if not match:
                retval.append(port_exp)
                continue

            start_num = match.group(1)
            end_num = match.group(2)

            for i in range(int(start_num), int(end_num) + 1):
                retval.append(i)
            
        self.options.nports_to_scan_per_job = len(retval)
        self.options.ports_to_scan = retval
    
    
    def get_triggers_to_test(self):
        return self.options.triggers_to_test
    

    def lock_object(self):
        self.lock.acquire()

    def unlock_object(self):
        self.lock.release()
    
  
    ############### PRIVATE METHODS ###################

    
    def usage(self):        
        print "portbunny <host1> [host2] ... [hostn] [options]"    
        print "by default all well known ports are scanned."    


    """
    (1) Validates all host-expressions contained in
    self.hosts

    (2) Resolves hostnames

    (3) Breaks down subnets given in CIDR-notation
    into single hosts.

    (4) Saves the result in self.hosts

    """

    def parse_host_expressions(self):

        if self.hosts == []:
            print "Please tell the bunny what to scan."            
            sys.exit(1)
        
        final_host_list = []
                
        for host in self.hosts:

            ip_match   = ip_re.match(host)
            host_match = hostname_re.match(host)
            
            
            if not ( ip_match or host_match ):
                print "Invalid host-expression: " + host
                sys.exit(1)                
                

            # If this expression is a hostname, resolve it.

            if host_match:

                try:
                    host = socket.gethostbyname(host)                    
                except socket.gaierror:
                    print "Error: Cannot resolve " + host
                    sys.exit(1)   

            
            # Handle IP-ranges
            elif host.find('/') != -1 :

                try:
                    # Create list of hosts from CIDR-expression
                    # and interate through each element of the
                    # list.
                    for h in CIDR(host):

                        # Discard x.x.x.255 and  x.x.x.0
                        if re.compile("\.255").search(h.dotted_address):
                            continue
                        if re.compile("\.0").search(h.dotted_address):
                            continue

                        # Add the element to the final_host_list
                        final_host_list.append(h.dotted_address)

                    continue
            
                except ValueError:

                    print "Error while parsing: " + host
                    sys.exit(1)

            
            final_host_list.append(host)
        
        self.hosts = final_host_list
        


    """
    Handle arguments to -p which is a whitespace-seperated list of
    single-ports and port-ranges and make them available as
    self.options.port_ranges.
    
    """
    def ports_option_parser(self, option, opt_str, value, parser):
        assert value is None
        done = 0
        value = []
        rargs = parser.rargs

        while rargs:
            arg = rargs[0]

            # Stop if we hit the next option
            if ((arg[:2] == "--" and len(arg) > 2) or
                (arg[:1] == "-" and len(arg) > 1 and arg[1] != "-")):
                break
            else:  

                # Validate that arg is indeed a port-expression
                if not re.compile('^(\d+)(-(\d+))*$').match(arg):
                    print "invalid port-expression: " + arg
                    sys.exit(1)

                value.append(arg)
                del rargs[0]
        
        setattr(parser.values, option.dest, value)      


    """
    Parse 'triggers_to_use' which will be accepted
    in the form 'TRIGGER_NAME1 ROUND1 TRIGGER_NAME2 ROUND2 ...'    
    """
    
    def triggers_to_test_parser(self, option, opt_str, value, parser):
        assert value is None
        done = 0
        value = []
        rargs = parser.rargs
        
        
        cur_trigger = None
        
        while rargs:
            arg = rargs[0]

            # Stop if we hit the next option
            if ((arg[:2] == "--" and len(arg) > 2) or
                (arg[:1] == "-" and len(arg) > 1 and arg[1] != "-")):
                break
            else:
               
                if cur_trigger is None:                                  
                    cur_trigger = []
                    cur_trigger.append(arg)
                else:
                    
                    cur_trigger.append(arg)
                    value.append(cur_trigger)
                    cur_trigger = None
                    
                del rargs[0]
                
        if cur_trigger:
            print "Error: arguments to -t are malformed"
            sys.exit(1)

        if not value:
            print "Error: -t takes at least 2 arguments"
            sys.exit(1)

        setattr(parser.values, option.dest, value)
