
import threading, atexit
from interface import PortList, stopParserThread
from UserEventHandler import stopUserEventHandler

def kill_ui_logic(ui_logic):
    # Up the finished-job-mutex so that the
    # UI-Logic wakes up.
    ui_logic.scanner.finished_job_mutex.release()


class UILogic(threading.Thread):

    def __init__(self, scanner, options_parser, services, event_handler):
       
        threading.Thread.__init__(self)
 
        self.scanner = scanner
        self.options_parser = options_parser
        self.services = services
        self.event_handler = event_handler

        self.nactive_trigger_jobs = 0
        self.nactive_flood_jobs = 0
        self.pending_trigger_jobs = []
        self.pending_flood_jobs = []
        self.running = True
        
        self.initialized_sem = threading.Semaphore(0)

        atexit.register(kill_ui_logic, self)
        

    """
    Adds intial trigger-jobs and then waits for
    jobs to complete and feeds the scanner new jobs.
    """

    def run(self):
         
        # (1) Create initial trigger-jobs.

        hosts_to_scan = self.options_parser.get_hosts_to_scan()
        self.options_parser.gen_list_of_single_ports()

        # we have now initialized
        self.initialized_sem.release()
        
        if not self.options_parser.options.only_discover:
            print "+++ Will scan " + str(self.options_parser.options.nports_to_scan_per_job),
            print "ports on " + str(len(hosts_to_scan)) + " hosts. +++"

        # Each host we want to scan has to be triggered first
        self.pending_trigger_jobs = hosts_to_scan
        self.create_trigger_jobs()
        


        # (2) Wait for finished jobs
        
        while self.running:
            
            target_mode_tuple = self.scanner.poll_for_finished_jobs()

            if not target_mode_tuple:
                # Empty tuples are returned when the logic
                # should terminate.
                return
                
                
            (target, mode) = target_mode_tuple

            # Handle errors
            if target == 'ERROR':
                print "scanner-module returned an error. Quitting."
                
                # Shut down the scanner
                stopParserThread(self.scanner.parser)
                stopUserEventHandler(self.event_handler)
                return
            
            
            """ Handle finished Jobs  """

            if mode == 'FLOOD':
                
                # A flood-job has finished: Output results,
                # delete the flood-jobs and execute the next
                # pending job

                self.output_flood_job_summary(target)
                self.scanner.free_scanjob(target, mode)
                
                self.nactive_flood_jobs = self.nactive_flood_jobs - 1
                
                # Submit a pending job
                if self.pending_flood_jobs != []:
                    self.create_flood_jobs()
                

            elif mode == 'TRIGGER':
                
                # A trigger-job has finished. Output results,
                # execute the next pending trigger-job.
                # If no more pending trigger-jobs exist,
                # start creating flood-jobs

                self.output_trigger_job_summary(target)
                host_is_up = self.scanner.is_trigger_job_up(target)
                
                self.nactive_trigger_jobs = self.nactive_trigger_jobs - 1
                
                if host_is_up and not self.options_parser.options.only_discover:
                    self.pending_flood_jobs.append(target)
                
                # self.scanner.free_scanjob(target, mode)
    
                # Execute a pending trigger-job if there is one.
                if self.pending_trigger_jobs != []:
                    self.create_trigger_jobs()
                elif self.nactive_trigger_jobs == 0:  
                    # No pending-trigger-jobs left, no active
                    # trigger-jobs left, execute flood-jobs
                    
                    banner_printed = False
                    nhosts = 0
                    i = 0
                    for trig_job in self.scanner.triggerJobs:
                        if self.scanner.triggerJobs[trig_job].triggers == []:
                            continue
                        
                        if not banner_printed:
                            print "+++ Trigger-Phase done. The following hosts are up: +++"
                            banner_printed = True

                        nhosts = nhosts + 1
                        print self.scanner.triggerJobs[trig_job].target,
                        
                        i = i + 1
                        if i == 3:
                            print ""
                            i = 0

                    if i != 0:
                        print "\n"
                    
                    if not banner_printed:
                        print "All hosts seem down."
                    else:
                        print str(nhosts) + " hosts total."

                    self.create_flood_jobs()

            # No jobs left, exit.
            if self.nactive_trigger_jobs + self.nactive_flood_jobs == 0:
                print "All done"
                # Shut down the scanner
                stopParserThread(self.scanner.parser)
                stopUserEventHandler(self.event_handler)
                return
                

    
    """
    creates and executes trigger-jobs until
    ntrigger_jobs_at_once has been reached.
    """
    
    def create_trigger_jobs(self):
        
   
        scanner = self.scanner 
        options_parser = self.options_parser 
        ntrigger_jobs_at_once = options_parser.options.ntrigger_jobs_at_once
        

        while ntrigger_jobs_at_once > self.nactive_trigger_jobs:
            
            if self.pending_trigger_jobs == []:
                # No more pending jobs, exit.
                return

            # Create next pending trigger-job

            host = self.pending_trigger_jobs.pop(0)
        
            scanner.createTriggerJob(host)
        

            # And tell scanner which triggers to use
        
            syn_added = False
            tcp_ports = []
            udp_added = False
            udp_ports = []

            scanner.triggerJobs[host].clearMethodsList()

            for (name, round) in options_parser.get_triggers_to_test():
                # print "trying trigger " + name + " "+ str(round) + " on " + host 
                if name == 'TCP_SYN':
                    if not syn_added:
                        scanner.triggerJobs[host].appendToMethodList(name)
                        syn_added = True
                    tcp_ports.append(round)                

                elif name == 'UDP':
                    if not udp_added:
                        scanner.triggerJobs[host].appendToMethodList(name)
                        udp_added = True
                    udp_ports.append(round)

                else:
                    # Handle one-round-triggers
                    scanner.triggerJobs[host].appendToMethodList(name)

            if syn_added:
                scanner.triggerJobs[host].tcpSynPorts(tcp_ports)
            if udp_added:
                scanner.triggerJobs[host].udpPorts(udp_ports)

            self.nactive_trigger_jobs = self.nactive_trigger_jobs + 1
            scanner.triggerJobs[host].execute()

    """
    create and execute flood-jobs until
    nflood_jobs_at_once has been reached.
    """

    def create_flood_jobs(self):
        
        nactive_flood_jobs = self.nactive_flood_jobs
        pending_flood_jobs = self.pending_flood_jobs
        scanner = self.scanner
        
        options_parser = self.options_parser
        nflood_jobs_at_once = options_parser.options.nflood_jobs_at_once

        while nflood_jobs_at_once > self.nactive_flood_jobs:
            
            if pending_flood_jobs == []:
                break

            target = pending_flood_jobs.pop(0)
            
            # create the flood-scan-job
            self.nactive_flood_jobs = self.nactive_flood_jobs + 1
            port_list = PortList(options_parser.options.port_ranges)
            
            scanner.createFloodJob(target, port_list, scanner.triggerJobs[target].triggers)

            if options_parser.options.create_graph:
                scanner.floodJobs[target].setEventsToReport('2')
            
            scanner.floodJobs[target].execute()
    
    

    def output_trigger_job_summary(self, hostname):
        
        if self.scanner.triggerJobs[hostname].triggers == []:
            return

        i = 0
        print "Best triggers for " + hostname + ": "
        print "===================================================="
        for trigger in self.scanner.triggerJobs[hostname].triggers:
            print trigger.method + " " + trigger.rnd + " ",
            i = i + 1
            if i == 3:
                print ""
                i = 0
        print "\n===================================================="
        


    """
    nicely print results with this not
    so nice code.
    """

    def output_flood_job_summary(self, hostname):
        scanner = self.scanner
        services = self.services
        options_parser = self.options_parser

        nports_to_scan_per_job = options_parser.options.nports_to_scan_per_job

        # Output graph-file if one was created

        if scanner.floodJobs[hostname].cwnd_list != []:
            f = open('./cwnd_log' + scanner.floodJobs[hostname].target + '.txt', "w")
            for cwnd in scanner.floodJobs[hostname].cwnd_list:
                f.write(cwnd + "\n")
            f.close()
        

        port_state_dict = {}
        port_state_dict["CLOSED"] = {}
        port_state_dict["FILTERED"] = {}
        port_state_dict["OPEN"] = {}

        
        print "Results for " + hostname
        print "===================================================="
        
        for res in scanner.floodJobs[hostname].results.keys():                                                        
            port_state_dict[scanner.floodJobs[hostname].results[res]][res] = 1

        # Output OPEN ports
        for p in port_state_dict["OPEN"].keys():

            service_name = "UNKNOWN"

            if int(p) in services:
                service_name = services[int(p)]

            print hostname + ":\t" + p + "\tOPEN\t\t" + service_name 
            

        # If the number of CLOSED ports is bigger than the number of filtered ports
        if len(port_state_dict["CLOSED"]) > len(port_state_dict["FILTERED"]):

            # Output the filtered ports
                        
            for p in port_state_dict["FILTERED"].keys():                            
                if not (p in port_state_dict["CLOSED"]) and not (p in port_state_dict["OPEN"]):

                    service_name = "UNKNOWN"
                    if int(p) in services:
                        service_name = services[int(p)]

                    print hostname + ":\t" + p + "\tFILTERED\t" + service_name 

            print "all other ports are CLOSED."
                                               
        else:

            # Output the CLOSED ports if there number is smaller or equal to
            # the number of filtered ports
                        
            for p in port_state_dict["CLOSED"].keys():

                service_name = "UNKNOWN"
                if int(p) in services:
                    service_name = services[int(p)]
                    
                print hostname + ":\t" + p + "\tCLOSED\t\t" + service_name
                        
            print "all other ports are FILTERED"

        print str(nports_to_scan_per_job) + " ports scanned."
        print "===================================================="
