#include "state.h"
#include "packet_batch.h"

#include "../scanner_ui/scanner_output_queue.h"
#include "../packet_submitter.h"
#include "../scan_jobs.h"
#include "../scanner_module.h"
#include "../timespec_utils.h"

#include "../trigger_state/tcp_syn_method.h"

#include <linux/module.h>
#include <asm/div64.h>

/**
    \addtogroup PacketBatch
    @{
*/



/**
  Constructor
    
*/

struct packet_batch *create_packet_batch(__be16 *ports,
					 unsigned int nports,
					 struct trigger_instance *trigger)
{	
	struct packet_batch *new_batch;

	/* Do not allow the creation of a null-batch */
	
	if(nports == 0 || !ports)
		return NULL;

	new_batch = kmalloc(sizeof(struct packet_batch), GFP_KERNEL);
	
	if(!new_batch)
		return NULL;	
	
	memset(new_batch, 0, sizeof(struct packet_batch));
	
	
	new_batch->port_indices = kmalloc(sizeof(__u16) * nports, GFP_KERNEL);
	new_batch->probes_time_sent = kmalloc(sizeof(struct timespec) * nports,
					      GFP_KERNEL);
	
	if( (!new_batch->port_indices) || (!new_batch->probes_time_sent)){
		kfree(new_batch);
		return NULL;
	}
	
	memcpy(new_batch->port_indices, ports, sizeof(__u16) * nports);
			
	new_batch->size_of_batch = nports;		
	new_batch->trigger_node = trigger;       		
	
	return new_batch;

}


/**
   Destructor
   
*/

void delete_packet_batch(struct packet_batch *this)
{
	
	if(!this)
		return;
	

	if(this->port_indices){
		
		/* make sure to deregister all pointers
		 * from port_results to this batch */
		
		int t;
		for(t = 0; t < this->size_of_batch; t++){
			struct port_result *presult = 
				this->ports_to_scan[this->port_indices[t]];
		
			if(presult->port_is_in_batch == this)
				presult->port_is_in_batch = NULL;
		
		}									
		
		kfree(this->port_indices);
	
	}
	
	if(this->probes_time_sent)
		kfree(this->probes_time_sent);

	if(this->trigger_rcv_time)
		kfree(this->trigger_rcv_time);
	
	del_trigger_instance(this->trigger_node);
	kfree(this);
}

/**
  Returns true if scanning 'port' was
  part of this packet-batch.

  Else, returns false

*/

boolean batch_handles_port(struct packet_batch *this, int port_index)
{
	int t;
	for(t = 0; t < this->size_of_batch; t++)
		if(this->port_indices[t] == port_index)
			return TRUE;
       

	return FALSE;
}


/**
  Returns a pointer to the active batch which
  handles a given port or NULL if no such batch
  exists.
*/

struct packet_batch *get_active_batch_by_port_index(struct flood_state_context *context,
						    int port_index)
{		
	if(context->ports_to_scan[port_index])
		return context->ports_to_scan[port_index]->port_is_in_batch;
 
	return NULL;
 
}


void send_packet_batch(struct packet_batch *this,
		       struct scan_job_t *job)
{	
		
	struct trigger_finding_method *trig_method;	
        int t;
	
	/* Iterate through all ports of this batch and
	 * send packets */
	
	
	for(t = 0; t < this->size_of_batch; t++){
	  
		send_tcp_syn_packet(job->addr, this->port_indices[t],
				    SRC_PORT, 0xffffffff);				
	
		this->probes_time_sent[t] = CURRENT_TIME;
	}
	
	trig_method = trigger_finding_methods[this->trigger_node->method_id];
	/* register batch_id with trigger */

	if(trig_method->register_batch_id)
		trig_method->register_batch_id(this->trigger_node, this->port_indices[0]);
	
	/* now append the trigger */
	
	trig_method->sender_func(job, this->trigger_node);		
	this->time_sent = CURRENT_TIME;	
	
}


/**
   Output results of packet-batch in the form specified
   in the architecture-paper.
*/

void batch_output_filtered_ports(struct packet_batch *batch,
				 struct scan_job_t *job)
{
	struct flood_state_context *context = job->state_context;
	
	int t;
	char num_buf[128];
	
			
	for(t = 0; t < batch->size_of_batch; t++){
				
	  
		if(!context->ports_to_scan[batch->port_indices[t]]->exists)
			continue;
		
		
		if((context->ports_to_scan[batch->port_indices[t]]->state == FILTERED)){
			/* Port is filtered */
			
			scanner_output_queue_add("RESULT PORT_STATE ");
			scanner_output_queue_add(job->addr_str);
			scanner_output_queue_add(" FLOOD");
			sprintf(num_buf, " %d ", batch->port_indices[t]);
			scanner_output_queue_add(num_buf);
			scanner_output_queue_add("FILTERED");
			scanner_output_queue_add("\n");


			
		}
	
	}
		
}


void batch_mark_unknown_as_filtered(struct packet_batch *this,
				    struct scan_job_t *job)
{
	struct flood_state_context *context =
		job->state_context;
	
	int t;

	for( t = 0; t < this->size_of_batch; t++){
		
		/* absurdity check */
		
		if(! context->ports_to_scan[ this->port_indices[t] ] ){
			scanner_output_queue_add("ERROR -1 Almost dereferenced a NULL-ptr");
			scanner_output_queue_add("in batch_mark_unkown_as_filtered\n");
			return;
		}

		if(! (context->ports_to_scan[this->port_indices[t]]->exists )){
			context->nports_scanned++;
			context->ports_to_scan[this->port_indices[t]]->exists = 1;
			context->ports_to_scan[this->port_indices[t]]->state = FILTERED;
			
			
		}
	}
}

struct timespec *get_time_sent_by_port(struct packet_batch *this,
				      int port)
{
	int t;
	
	if(!this) return NULL;

	for(t = 0; t < this->size_of_batch; t++){
		if(this->port_indices[t] == port)
			return &this->probes_time_sent[t];
	}
	
	return NULL;
}

boolean batch_timed_out(struct packet_batch *this,
			s64 cur_timeout,
			struct flood_state_context *context)
{
	struct timespec cur_time = CURRENT_TIME;
	struct timespec timeout_time = this->time_sent;	

	boolean timed_out = FALSE;	
	
	timespec_add_ns(&timeout_time, cur_timeout);
	
	timed_out = (timespec_compare(&timeout_time, &cur_time) <= 0);
	
	if(timed_out){
		printk("Hit clock\n");
		return TRUE;
	
	}

	/* check if any batch has not received its trigger-response
	   and has a sequence-number smaller than 
	 * the biggest sequence_number received */

	
	if ((!this->trigger_rcv_time) &&
	    ( this->seq_num + 3 < context->max_batch_seq_received)){
		printk("fast-retransmit\n");
		return TRUE;
	}
	
	
	return FALSE;
	
}


/** @} */
