
#include "cmd_handlers.h"
#include "../scanner_module.h"
#include "../scan_jobs.h"
#include "../scan_job_states.h"
#include "../scanner_ui/cmd_handlers.h"
#include "../scanner_ui/scanner_output_queue.h"

#include "../scanner_module.h"
#include "../packet_submitter.h"


#include "../trigger_state/trigger_finding_methods.h"
#include "state.h"
#include "packet_batch.h"
#include "batch_creator.h"
#include "event_reporters.h"
#include "tcp_reno.h"

#include <asm/div64.h>
#include <linux/vmalloc.h>

#include "../timespec_utils.h"

/** \addtogroup FloodStateScanJobManager
@{
*/

static struct cmd_handlers_hash_bucket cmd_handlers_hash[CMD_HASH_SIZE];

/**
   Global congestion window.
   
   This window is used by all scan-jobs. Only if the
   scan-jobs own window and the global window allow
   the scan-job to send new packets, will it do so.  

   Note: This window is currently not respected by
   send_new_packet_batches because it turns out to
   be not the way to go.
   
*/

struct flood_state_timing_context global_timing_context ={
	.cur_cwnd = INITIAL_GLOBAL_CWND,
	.cur_ccthresh = INITIAL_GLOBAL_CCTHRESH,
	.no_responses_since_drop = FALSE,
	.nactive_packets = 0,
};


/**
   FloodStateScanJob-Constructor
   
   TODO: trigger-list does not make use of the generic
   queue-class yet and is not freed in the scan-job-destructor

*/


static int scan_job_init(struct scan_job_t *this)
{
	struct flood_state_context *context;
	unsigned int t = 0;

	
	/* Allocate memory for state-context and initialize it to 0. */

	context = this->state_context
		= kmalloc(sizeof(struct flood_state_context), GFP_KERNEL);

	if(!this->state_context){
		this->state_context = NULL;
		return FAILURE;
	}

	memset(this->state_context, 0, sizeof(struct flood_state_context));
	
	
	/* Initialize timing */
	
	context->timing_alg = &tcp_reno;
	context->timing_algo_state =
		context->timing_alg->constructor(&context->timing_context);


	/* Initialize default list of triggers
	   to an empty list */
	
	INIT_LIST_HEAD(&context->triggers.list);			
	
	/* Initialize lists used for batch-management */	
	context->active_batches = create_queue(NULL, GFP_KERNEL, FALSE);
	context->recent_batches = create_queue(NULL, GFP_KERNEL, FALSE);
	

	if(!context->active_batches || !context->recent_batches){
		kfree(this->state_context);
		this->state_context = NULL;
		return FAILURE;
	}
		
	/* initialize the packet-queue used for incoming packets */			
	
	INIT_LIST_HEAD(&this->packets.list);
	this->packets.lock = kmalloc(sizeof(rwlock_t), GFP_KERNEL);
	
	if(!this->packets.lock){
		delete_queue(context->active_batches, NULL);
		delete_queue(context->recent_batches, NULL);
		kfree(this->state_context);
		this->state_context = NULL;
		return FAILURE;
	}
	
	*this->packets.lock = RW_LOCK_UNLOCKED;
	
	/* initialize ports to scan to a valid value
	 * just to make sure people don't crash their kernel
	 * simply because they forgot to use set_ports_to_scan
	 */
	
	context->nports_to_scan = NDEFAULT_PORTS_TO_SCAN;
	context->port_array_size = NDEFAULT_PORTS_TO_SCAN + 1;
	context->ports_to_scan = vmalloc(sizeof(struct port_result *) * context->port_array_size);
	
	if(!context->ports_to_scan){
		kfree(this->packets.lock);
		delete_queue(context->active_batches, NULL);
		delete_queue(context->recent_batches, NULL);
		kfree(this->state_context);
		this->state_context = NULL;
		return FAILURE;
	}

	memset(context->ports_to_scan, 0, sizeof(struct port_result *) * context->port_array_size);	
	
	for(t = 1; t <= NDEFAULT_PORTS_TO_SCAN; t++){
		context->ports_to_scan[t] = create_port_result(t);
	}

	return SUCCESS;

}


static void clear_trigger_list(struct scan_job_t *this)
{
	struct flood_state_context *context
		= this->state_context;

	struct list_head *p, *n;
	list_for_each_safe(p, n, &context->triggers.list){
		struct trigger_instance *node		
			= list_entry(p, struct trigger_instance, list);
		
		list_del(p);
		del_trigger_instance(node);
	}
	
}

/**
        Flood-state-scan-job Destructor
*/

static void scan_job_fini(struct scan_job_t *this)
{
	
	struct flood_state_context *context
		= this->state_context;
		
	
	/* delete queues holding batches */
	
	delete_queue(context->active_batches, (delete_data_func) delete_packet_batch);
	delete_queue(context->recent_batches, (delete_data_func) delete_packet_batch);
	
	/* delete list of triggers */
	clear_trigger_list(this);
	

	if(context->creator)
		delete_batch_creator(context->creator);
		
	
	/* free old ports_to_scan array if any. */

	if(context->ports_to_scan){
		
		int t;
		for(t = 0; t < context->port_array_size; t++)
			if(context->ports_to_scan[t])
				delete_port_result(context->ports_to_scan[t]);
				
		vfree(context->ports_to_scan);		
	}
		
	
	if(this->packets.lock)
		kfree(this->packets.lock);

	if(this->addr_str)
		kfree(this->addr_str);
	
	context->timing_alg->destructor(context->timing_algo_state,
					&context->timing_context);
	
	kfree(this->state_context);
	kfree(this);
}

/**************************************************
  Processing of incoming packets in flood-state
***************************************************/

/**

Called for all possible probe-responses.

Registers the port_result indicated by the ACK or RST.

Requirements: descr is a packet-description which has been
identified to describe a tcp-syn-ack-packet or a tcp-rst-packet.

*/

static void process_tcp_ack_or_rst(struct scan_job_t *this,
				   struct sniffed_packet_descr_t *descr)
{
	struct flood_state_context *context = this->state_context;	
	int port = descr->port;
	struct port_result *result;
	struct trigger_instance *node;		
	struct packet_batch *batch;

	
	/* 1. Is the port being scanned at all? */

	/* check if port is between 0 and biggest port-number scanned */

	if(!((port > 0) && (port < context->port_array_size))){
		printk("port-number %d out of range\n", port);
		return;
	}

	result = context->ports_to_scan[port];
	
	if(!result){
		printk("port-number %d not being scanned\n", port);
		return;
	}
	
	
	/* Check if there is already a result registered for
	 * this port which was created using an
	 * earlier response.
	 */
	
	if( (result->exists) && (result->state != FILTERED) ){		
		printk("port-result %d already known\n", port);						
		return;
	}
	
	
	/* 2. Is this probe currently in an active batch
	 * or in the list of recent batches?
	 */
	
	batch = result->port_is_in_batch;
	
	if(!batch){
		printk("tcp-ack/rst has valid form but it can't be found");
		printk(" in active or recent batches\n");
		
		if(descr->tcp_rst)
			printk("tcp_rst\n");

		return;	
	}
	
	/* Yes, so use it for timeout-calculation */
	
	context->timing_alg->on_update_timeout(context->timing_algo_state,
					       &context->timing_context,
					       get_time_sent_by_port(batch, port),
					       descr);
	
		
	/* register result */	
	
	/* given that result does not yet exist,
	 * increase nports_scanned. If this port
	 * is currently in a batch
	 */

	if(!result->exists)
		context->nports_scanned++;		
	
	result->exists = TRUE;	
	//printk("registering result for port: %d\n", port);
	
	/* output new port-state */
		
	if(descr->tcp_syn_ack || descr->tcp_rst || descr->tcp_ack_rst){
		char num_buf[128];
				
			
		scanner_output_queue_add("RESULT PORT_STATE ");
		scanner_output_queue_add(this->addr_str);
		scanner_output_queue_add(" FLOOD");
		sprintf(num_buf, " %d ", port);
		scanner_output_queue_add(num_buf);
		
		if(descr->tcp_syn_ack){
			result->state = OPEN;	
			scanner_output_queue_add("OPEN");
		}else{
			
			if(descr->icmp_code == ICMP_PKT_FILTERED){
				
				result->state = FILTERED;
				scanner_output_queue_add("FILTERED");								
				
			}else{
							
				result->state = CLOSED;
				scanner_output_queue_add("CLOSED");				
				
			}
		
			output_negative_response(this, descr);		
		}
		
		scanner_output_queue_add("\n");		
	
	}		
	
	
	/* use this port as a new trigger if the maximum
	 * number of triggers has not been reached. */	

	
	if(descr->icmp_code != ICMP_PKT_FILTERED){
		
		if( ++ context->nextra_triggers > MAX_EXTRA_TRIGGERS){
			context->nextra_triggers--;
			return;
		}
		

		node = new_trigger_instance(TCP_SYN_FINDING_METHOD, port);
		
		if(!node){
			printk("Error creating new trigger-node\n");
			return;
		}
		
		list_add_tail(&node->list, &context->triggers.list);	
		output_trigger_added_event(this, port);
	}
	

		
}


/**
   Determine the trigger-type (trigger-method) of
   the trigger described by descr and the batch
   this trigger belongs to.
   
   Register the trigger-response.  
   
*/

static void process_trigger_responses(struct scan_job_t *this,
				      struct sniffed_packet_descr_t *descr)
{
	struct flood_state_context *context = this->state_context;	
	struct list_head *p;
	
	
	/* iterate all triggers in use to see if one matches:
	 * This loop is another reason to limit the maximum
	 * number of extra triggers to use.
	 */

	list_for_each(p, &context->triggers.list){
		struct trigger_instance *entry = 
			list_entry(p, struct trigger_instance, list);
		int method_id = entry->method_id;
		
		int batch_id = 
			trigger_finding_methods[method_id]->extract_batch_id
			((struct sniffed_packet_descr_t *) descr);
		
		
		struct packet_batch *batch = NULL;
		struct port_result *result;

		/* continue if this must have been another trigger */

		if(batch_id == -1 || batch_id == 0)
			continue;

		if(! ((batch_id > 0 ) && (batch_id < context->port_array_size))){
			printk("batch_id out of range\n");
			return;
		}		

		
		result = context->ports_to_scan[batch_id];		
		
		if(!result){
			
			/* This can't be a response to our trigger because
			 * a trigger with this seq-num will never and has
			 * never been sent.*/
			
			printk("invalid trigger-sequence-number\n");
			return;
		}

		batch = result->port_is_in_batch;
		
		if(!batch){
			printk("packet has correct trigger-response-form but");
			printk(" cannot be found in active or recent batches\n");
			return;
		}


		if(batch->trigger_rcv_time){
			printk("Trigger %d already has a response\n", batch->port_indices[0]);
			return;
		}
		
		

		/* Handle late trigger-responses */
		
		if(batch->inactive){
			
			printk("late-trigger response received\n");
			
			/* rewind cwnd */
			
			context->timing_alg->on_late_response(context->timing_algo_state,
							      &context->timing_context
				);				
			
			context->timing_alg->on_late_response(NULL,
							      &global_timing_context
				);
			

			/* and act as if the batch completed normally. */

			batch_mark_unknown_as_filtered(batch, this);
			batch_output_filtered_ports(batch, this);


		}

		
		/* write down the trigger-receive time */
		batch->trigger_rcv_time = kmalloc(sizeof(struct timespec), GFP_KERNEL);
		if(!batch->trigger_rcv_time)
			return;
		
		/* trigger response received */				

		memcpy(batch->trigger_rcv_time, &descr->time_received, sizeof(struct timespec));				
		
		if(context->max_batch_seq_received < batch->seq_num)
			context->max_batch_seq_received = batch->seq_num;

		/* report trigger-event */
		
		if(context->events_to_report & TRIGGER_EVENT){
			int round =
				trigger_finding_methods[method_id]->get_round_by_descr(descr);
			
			output_trigger_received_event(this, method_id, round);			
		}
		
		/* if this trigger triggered a tcp_syn_ack, write this
		 * down! This may result in ports being reported OPEN
		 * twice but at least we wont miss the port.
		 */
		
		if(descr->tcp_syn_ack){
			struct port_result *result = 
				context->ports_to_scan[descr->port];						
			
			if(!result) return;

			if(result->state != OPEN){
				char num_buf[128];
	

				result->state = OPEN;	
				
				/* output OPEN ports. */
				
				scanner_output_queue_add("RESULT PORT_STATE ");
				scanner_output_queue_add(this->addr_str);
				scanner_output_queue_add(" FLOOD");
				sprintf(num_buf, " %d ", descr->port);
				scanner_output_queue_add(num_buf);
				scanner_output_queue_add("OPEN");
				scanner_output_queue_add("\n");
				
			}	
		}


		/* use trigger-rtt to update timeout-value */

		context->timing_alg->on_update_timeout(context->timing_algo_state,
						       &context->timing_context,
						       &batch->time_sent,
						       descr);		
		output_timeout_info(this);
		
		return;
		

	}

}



/**
   Polls the sniffed_packet_queue of this ScanJob which
   contains all newly arrived packets.

   For each newly arrived packet, it checks whether this
   is a trigger-response or a probe-response. In the
   case of a probe-response, \ref process_tcp_ack_or_rst is called.
   Else, \ref process_trigger_response is called.
   
*/


static void process_pending_packets(struct scan_job_t *this)
{
	struct sniffed_packet_descr_t *p_descr;
	struct flood_state_context *context = this->state_context;

	while((p_descr = sniffed_packet_queue_head(&this->packets))){
		
		struct sniffed_packet_descr_t *descr =
			(struct sniffed_packet_descr_t *) p_descr;		
		
		struct timespec cur_time = CURRENT_TIME;	
		
		memcpy(&descr->time_received, &cur_time, sizeof(struct timespec));		

		/* Hack: we want to treat ICMP adm-filtered packets just like
		 * TCP_ACK_RSTs with the exception that the port-result
		 * is filtered instead of closed.
		 *
		 * So what we do is convert the descr to a TCP_SYN_ACK-packet
		 * and we keep the icmp_code at 13 (admin-prohib.).
		 */

		if(descr->protocol == IPPROTO_ICMP &&
		   descr->icmp_type == ICMP_DEST_UNREACH &&
		   descr->icmp_code == ICMP_PKT_FILTERED){
			
			descr->protocol = IPPROTO_TCP;
			descr->ack_seq = 0;
			descr->tcp_ack_rst = 1;
			

		}
							


		/* If this is a tcp-syn-ack with an ack_seq of 0,
		   process it as a reply to a scan-probe as opposed
		   to a trigger.
		*/
		      	

		if( (descr->protocol == IPPROTO_TCP) &&
		    (descr->src == this->addr)       &&	    
		    ( descr->tcp_rst || descr->tcp_syn_ack || descr->tcp_ack_rst ) &&
		    (descr->ack_seq == 0)) 						
			process_tcp_ack_or_rst(this, descr);
		
		else		  						
			process_trigger_responses(this, descr);
		
				
		
		if(context->timing_context.no_responses_since_drop)
			context->timing_context.no_responses_since_drop = FALSE;
				
		kfree(descr);
	}
	
}

static boolean all_port_states_known(struct scan_job_t *this)
{
	struct flood_state_context *context
		= this->state_context;


	return (context->nports_scanned == context->nports_to_scan);

}


/**************************************************
   Sending new packets:
****************************************************/


/* Warning: We're currently not taking into account
 * that as we have fewer ports available to scan, we might
 * be thinking that we have actually sent more than we did.
 * This is of course not good for cwnd-calculation.
 */


static void send_new_packet_batches(struct scan_job_t *this)
{
	struct flood_state_context *context
		= this->state_context;

	struct batch_creator *creator
		= context->creator;

	unsigned int ncomplete_batches;
	unsigned int nextra_probes;
	struct packet_batch **new_batches;
	unsigned int t;
	unsigned int nbatches = 0;
	//s64 global_nnew_packets;
	s64 nnew_packets;

	/* Calculate number of packets to send:
	 * current_window - nactive_packets
	 * 
	 * This isn't pleasant to look at because
	 * we're using s64 and unsigned int +
	 * a blowup-factor since floating-point-
	 * arithmetic in the kernel isn't possible.
	 */
	
	s64 nactive_packets = context->timing_context.nactive_packets;
	nactive_packets *= CWND_BLOWUP_FACTOR;
	
	
	/* If there are currently more active packets than allowed,
	   do not send new ones.*/
	
	if(context->timing_context.cur_cwnd < nactive_packets)
		return;

	nnew_packets = 
		context->timing_context.cur_cwnd - nactive_packets;	
	
	do_div(nnew_packets, CWND_BLOWUP_FACTOR);
	

	/* Note: The global window is no longer in effect.
	 * It's not a good idea at all because you end up
	 * letting all scan-jobs perform like the worst
	 * scan-job in the group.
	 */

	/* Take into account the global window.
	 * If the global window does not allow
	 * us to send the number of probes we want to
	 * send, take what you can get.
	 */
	
	//nactive_packets = global_timing_context.nactive_packets;
	//nactive_packets *= CWND_BLOWUP_FACTOR;
	
	/* If there are currently more active packets than allowed,
	   do not send new ones.*/

	
	//if(global_timing_context.cur_cwnd < nactive_packets)
	//	return;
	

	//global_nnew_packets = global_timing_context.cur_cwnd - 
	//	nactive_packets;
	//do_div(global_nnew_packets, CWND_BLOWUP_FACTOR);
		
	//if(global_nnew_packets < nnew_packets)
	//	nnew_packets = global_nnew_packets;
	


	/* One may never send a single probe
	 * without a trigger. So 2 packets is
	 * the absolute minimum that needs to
	 * be sent at once.
	 */
		
	if(nnew_packets < 2) return;
		

	/* create new batches: */

	ncomplete_batches = 
		((unsigned int) nnew_packets) / (creator->cur_batch_size + 1);
	nextra_probes = nnew_packets - 
		ncomplete_batches * (creator->cur_batch_size + 1);
			

	new_batches =  kmalloc(sizeof(struct packet_batch *) * (ncomplete_batches + 1),
			       GFP_KERNEL);

	if(!new_batches) return;
	
	memset(new_batches, 0, sizeof(struct packet_batch *) * (ncomplete_batches + 1) );
	
	/* (1) create complete batches */
	
	for(t = 0; t < ncomplete_batches; t++){
		
		new_batches[t] =
			batch_creator_create_batch(creator);
		
		if(!new_batches[t]){
			
			/* no further batch could be created.
			 * We're probably out of ports to scan
			 */
			
			break;
		}
								
		nbatches++;			

	}			

	
	if( (nbatches == 0) && (ncomplete_batches > 0) ){
		/* Not a single batch could be created... */				
		kfree(new_batches);		
		return;
	}

	

	/* If there are extra probes or there is only one batch
	   to be created, create last or one and only batch */
	
	if( ! ((nbatches != ncomplete_batches) || (nextra_probes <= 1) )){
		
		
		
		/* create last batch or, if npackets_requested < cur_batch_size + 1,
		 * create the one and only batch.
		 */
		
		unsigned int old_batch_size = creator->cur_batch_size;		
				
		batch_creator_set_cur_batch_size(creator, nextra_probes - 1);				
		new_batches[ncomplete_batches] = batch_creator_create_batch(creator);
		batch_creator_set_cur_batch_size(creator, old_batch_size);


		if(!new_batches[ncomplete_batches] && (ncomplete_batches == 0)){
			
			/* One and only batch could not be created. */

			kfree(new_batches);			
			return;
			
		}
		
	}
	
	/* now add all batches created */
	
	for(t = 0; t < (ncomplete_batches + 1); t++){
		if(new_batches[t]){
			queue_add(context->active_batches, new_batches[t]);
			send_packet_batch(new_batches[t], this);
			
			/* update local and global number of active packets */
			context->timing_context.nactive_packets += new_batches[t]->size_of_batch + 1;
			global_timing_context.nactive_packets += new_batches[t]->size_of_batch + 1;
		}
	}


	kfree(new_batches);

}

static void manage_batches(struct scan_job_t *this)
{
	struct flood_state_context *context = 
		this->state_context;
	
	/* for each active burst, check if it has finished
	 * or timed out.*/
	struct list_head *p, *n;
	struct queue_node_t *root = get_root(context->active_batches);
	
	/* for each active batch */
	
	list_for_each_safe(p, n, &root->list){
		struct queue_node_t *entry = 
			list_entry(p, struct queue_node_t, list);
		struct packet_batch *batch = entry->data;

		if(batch->trigger_rcv_time){
			unsigned int t;
			
			
			/* trigger has been received, increase cwnd accordingly */

			context->timing_alg->on_rcv(context->timing_algo_state,
						    &context->timing_context);
			
			context->timing_alg->on_rcv(NULL,
						    &global_timing_context);


			/* trigger is no longer counted as active */
			context->timing_context.nactive_packets--;
			global_timing_context.nactive_packets--;
			
			/* port-states which are still unknown can now
			 * be marked as filtered.
			 */

			batch_mark_unknown_as_filtered(batch, this);
			
			/* all probes of this batch are now not active anymore:
			 * reduce nactive_packets and increase cwnd accordingly.
			 */
			
			for(t = 0; t < batch->size_of_batch; t++){										
				
				/* increase cwnd for each probe in the batch */
				
				context->timing_alg->on_rcv(context->timing_algo_state,
						    &context->timing_context);
				
				context->timing_alg->on_rcv(NULL,
							    &global_timing_context);
				
				
				/* decrease number of active packets for each probe of the batch */
				context->timing_context.nactive_packets--;
				global_timing_context.nactive_packets--;
			}			
			
						
			/* remove batch from list */
                        
			list_del(&entry->list);                 
                        kfree(entry);
                        
			batch_output_filtered_ports(batch, this);		
			
			
			/* save batch in recent-batches */
			
			batch->inactive = TRUE;
			queue_add_limited(context->recent_batches, batch,
					  (delete_data_func) delete_packet_batch,
					  NUM_OLD_BATCHES_TO_STORE);
			
					  

			output_cwnd_updated_event(this);


		}
		else if(batch_timed_out(batch, context->timing_context.cur_timeout, context)){
			int t;

			
			/* batch has timed out.
			 * Reinsert any unknown ports
			 * from timed-out batch into
			 * the batch-creator so that they
			 * can be scanned again.
			 */
					
			
			if(! context->timing_context.no_responses_since_drop){
				context->timing_alg->on_drop(context->timing_algo_state,
							     &context->timing_context,
							     NULL);
				
				context->timing_alg->on_drop(NULL,
							     &global_timing_context,
							     NULL);

				output_cwnd_updated_event(this);	
			}

																		
			/* trigger is no longer counted as active */
			context->timing_context.nactive_packets--;
			global_timing_context.nactive_packets--;
			
			if(context->events_to_report & TRIGGER_EVENT){						
				int round = trigger_instance_get_round(batch->trigger_node);
				output_dropped_trigger_event(this, batch->trigger_node->method_id, round);
			}
				
										
			/* all probes of this batch are now not active anymore */
			
			for(t = 0; t < batch->size_of_batch; t++){										
				context->timing_context.nactive_packets--;
				global_timing_context.nactive_packets--;
			}								

			
			batch_creator_reinsert_ports_of_batch(context->creator, batch);
			
			context->timing_context.no_responses_since_drop = TRUE;			
			
			/* remove batch from queue of active-batches */
			list_del(&entry->list);			
			kfree(entry);
		       	
						
			/* save batch in recent_batches */
			
			batch->inactive = TRUE;
			queue_add_limited(context->recent_batches, batch,
					  (delete_data_func) delete_packet_batch,
					  NUM_OLD_BATCHES_TO_STORE);
			
		}
		
		
	}
	


}


static void rescan_filtered(struct scan_job_t *this)
{
	struct flood_state_context *context = this->state_context;
	
	unsigned int nfiltered = 0;
	unsigned int nothers = 0;
	int t;

	/* count number of filtered/not filtered ports */

	for(t = 0; t < context->port_array_size; t++){
		
		if(!context->ports_to_scan[t])
			continue;

		if(context->ports_to_scan[t]->state == FILTERED)
			nfiltered++;
		else
			nothers++;

	}

	/* if the number of filtered ports is smaller bigger than 30%,
	 * don't rescan.
	 */
	
	if ((nfiltered * 100)/(nfiltered + nothers) > 30)
		return;
	
	
	/* rescan */

	for(t = 0; t < context->port_array_size; t++){
		
	
		if(context->ports_to_scan[t] && context->ports_to_scan[t]->state == FILTERED){
			
			/* mark port-states as unknown */
			context->ports_to_scan[t]->exists = FALSE;
			context->ports_to_scan[t]->port_is_in_batch = NULL;
			

			/* add port to batch-creator */
			batch_creator_add_port(context->creator, context->ports_to_scan[t]);		
			context->nports_scanned--;
			
		}

	}


}

/**
   FloodStateScanJobManager-Function.
   
   When initialy called, a \ref BatchCreator
   is created and initialized with the
   ports_to_scan-array.
   
   The ScanJobManagerFunction then does
   the following each round:
        
   (1) it processes any pending packets
   by calling
   \ref FloodStateScanJobManager::process_pending_packets.
   
   (2) It manages active \ref Burst s by
   calling \ref manage_bursts.

   (3) It outputs any finished batches by
   calling \ref output_finished_batches.

   (4) It checks if all port states are now
   known and we are finished waiting for late
   responses. If so, it returns FINISHED.
   
   (5) It sends any new batches by calling
   \ref send_new_packet_batches


*/

static int flood_state_scan_job_manager(struct scan_job_t *this)
{
	struct flood_state_context *context = this->state_context;		
	
	

	if(!context->initialized){
	
		/* Initialization */
	
		context->creator = create_batch_creator(context->ports_to_scan,
							context->port_array_size,
							&context->triggers);
		
		batch_creator_set_cur_batch_size(context->creator,
						 context->timing_context.cur_batch_size);
		
		context->initialized = TRUE;			
		
	}
	
	if(list_empty(&context->triggers.list)){
		scanner_output_queue_add("ERROR -1 No triggers available for flood-job\n");
		return FINISHED;
	}
	

	process_pending_packets(this);				
	manage_batches(this);	
	
	/* We're done once we know all port-states
	 * and we have no more batches waiting for
	 * late packets in the complete_batches-queue.
	 */

	if(all_port_states_known(this)){
		
		/* rescan filtered ports if necessary */
		
		if(context->nrescanned_filtered != NRESCANS_OF_FILTERED_PORTS){
			rescan_filtered(this);
			context->nrescanned_filtered++;
			printk("%d rescans of filtered ports performed.\n", context->nrescanned_filtered);
		}else						
			return FINISHED;			
			
		
	}
	
	
	send_new_packet_batches(this);		

	return CALL_AGAIN;
	
}


static int flood_state_init(void)
{
		
	cmd_handlers_init(flood_state.cmd_handlers);

	/* now register all cmd-handlers */
	
	cmd_handlers_register(flood_state.cmd_handlers,
			      "set_ports_to_scan",
			      &set_ports_to_scan_handler);

	cmd_handlers_register(flood_state.cmd_handlers,
			      "append_to_trigger_list",
			      &append_to_trigger_list_handler);
	
	
	cmd_handlers_register(flood_state.cmd_handlers,
			      "clear_trigger_list",
			      &clear_trigger_list_handler);

	cmd_handlers_register(flood_state.cmd_handlers,
			      "set_events_to_report",
			      &set_events_to_report_handler);
	

	return SUCCESS;
}


static void flood_state_fini(void)
{
	cmd_handlers_fini(flood_state.cmd_handlers);
}


/**
   scan_job_state_t-implementation exported
   and saved in \ref scan_job_states so that
   the ScanJobManager can make use of this state.
   
*/

struct scan_job_state_t flood_state = {
		
	.name = "FLOOD",	
	.scan_job_manager = &flood_state_scan_job_manager,
	.scan_job_init = &scan_job_init,
	.scan_job_fini = &scan_job_fini,
	.init = &flood_state_init,
	.fini = &flood_state_fini,
	
	.cmd_handlers = cmd_handlers_hash


};

/** @} */
