#include "tcp_reno.h"
#include "timing.h"
#include "state.h"

#include "../timespec_utils.h"

#include <asm/div64.h>

/*
  Implementation of TCP-Reno for PortBunny
*/

static void quick_start_rcv(struct flood_state_timing_context *tcontext)
{	
	
	/* Got an answer , increase cwnd */
	/* cwnd += 1 probe */
	
	tcontext->cur_cwnd += CWND_BLOWUP_FACTOR;					
	
}

static void congestion_avoidance_rcv(struct flood_state_timing_context *tcontext)
{
				
	/* Got an answer. */
	
	/* cwnd += 1/cwnd */
				
	s64 cur_cwnd = tcontext->cur_cwnd;
	s64 cwnd_blowup = CWND_BLOWUP_FACTOR;
		
	do_div(cur_cwnd, CWND_BLOWUP_FACTOR);
	do_div(cwnd_blowup, cur_cwnd);
	
	tcontext->cur_cwnd += cwnd_blowup;
		
}


static void update_timeout(struct tcp_reno_state *this,
			   struct flood_state_timing_context *tcontext,
			   struct timespec *time_sent,
			   struct sniffed_packet_descr_t *pdescr)
{
				
	
	const unsigned int AVERAGE_GAIN_RECIPROCAL = 3;
	const unsigned int DEVIATION_GAIN_RECIPROCAL = 2;
	
	struct timespec rtt;
	struct timespec null_time=  { .tv_sec = 0, .tv_nsec = 0, };
	s64 rtt_nsec;
	s64 rtt_error;	
	
	if(timespec_compare(&pdescr->time_received, time_sent) <= 0)
		rtt = null_time;
	else			
		rtt = timespec_sub( pdescr->time_received,
				    *time_sent );
	
	//printk("rtt: %ld, %ld\n", rtt.tv_sec, rtt.tv_nsec);

	
	if(rtt.tv_sec >= 5){
		printk("rtt bigger than 5 seconds encountered, discarding\n");
		return;
	}
				
	/* calculate new average round-trip-time ('rtt') */
	
	rtt_nsec = timespec_to_ns( &rtt );
	rtt_error = rtt_nsec - (this->rtt_average >> AVERAGE_GAIN_RECIPROCAL);					
	this->rtt_average += rtt_error;
	
	/* calculate new trigger-deviation */
	
	if(rtt_error < 0)
		rtt_error = -rtt_error;
	
	rtt_error -= (this->rtt_deviation >> DEVIATION_GAIN_RECIPROCAL);
	this->rtt_deviation += rtt_error;
	
	/* update the timeout to be 4*(average + 4 * deviation) */
	
	tcontext->cur_timeout = 4 * ((this->rtt_average >> 3)
				     + this->rtt_deviation );
				     
	
	if(tcontext->cur_timeout < MINIMUM_BATCH_TIMEOUT_NS )
		tcontext->cur_timeout = MINIMUM_BATCH_TIMEOUT_NS;
	
	
}


/********************************************/

static void reno_on_rcv(void *this,
			struct flood_state_timing_context *tcontext)

{	
        /* Depending on the current cwnd and threshold, determine
	 * what congestion-control-state we are in.
	 */
	
	if(tcontext->cur_cwnd < tcontext->cur_ccthresh)
		quick_start_rcv(tcontext);
	else
		congestion_avoidance_rcv(tcontext);		
		
}

static void reno_on_update_timeout(void *this,
				   struct flood_state_timing_context *tcontext,
				   struct timespec *time_sent,
				   struct sniffed_packet_descr_t *pdescr)
{

	if(!this) return;

	if(time_sent && pdescr)
		update_timeout(this, tcontext, time_sent, pdescr);
	
	
}

static void reno_on_drop(void *unused,
			 struct flood_state_timing_context *tcontext,
			 struct sniffed_packet_descr_t *pdescr)
{
	/* Got no answer, decrease cwnd and ccthresh */		
	
	tcontext->cwnd_before_drop = tcontext->cur_cwnd;
	tcontext->ccthresh_before_drop = tcontext->cur_ccthresh;       
	
	tcontext->cur_ccthresh = (tcontext->nactive_packets / 2) * CWND_BLOWUP_FACTOR;		
		
	do_div(tcontext->cur_cwnd,2);
	
	if(tcontext->cur_cwnd < 2*CWND_BLOWUP_FACTOR)		 
		tcontext->cur_cwnd = 2 * CWND_BLOWUP_FACTOR;
	
	if(tcontext->cur_ccthresh < 2*CWND_BLOWUP_FACTOR)
		tcontext->cur_ccthresh = 2*CWND_BLOWUP_FACTOR;				
	
}



static void *reno_constructor(struct flood_state_timing_context *tcontext)
{
	struct tcp_reno_state *this = kmalloc(sizeof(struct tcp_reno_state),
					      GFP_KERNEL);
	
	if(!this) return NULL;

	this->rtt_average = DEFAULT_BATCH_TIMEOUT_NS;
	this->rtt_deviation = 0;
	
	
	/* Now initialize timing-context */

	tcontext->cur_timeout = DEFAULT_BATCH_TIMEOUT_NS;	
	tcontext->cur_batch_size = DEFAULT_BATCH_SIZE;
	
	tcontext->cur_cwnd = INITIAL_CWND;	
	tcontext->cur_ccthresh = INITIAL_CCTHRESH;	

	return this;
	
}

static void reno_destructor(void *this,
			    struct flood_state_timing_context *tcontext)
{
	if(this)
		kfree(this);
	
}

static void reno_on_late_response(void *unused,
				  struct flood_state_timing_context *tcontext)
{			
	
	tcontext->cur_cwnd = tcontext->cwnd_before_drop;
	tcontext->cur_ccthresh = tcontext->ccthresh_before_drop;		
	
	if(tcontext->cur_cwnd < 2 * CWND_BLOWUP_FACTOR)
		tcontext->cur_cwnd = 2 * CWND_BLOWUP_FACTOR;
	if(tcontext->cur_ccthresh < 2 * CWND_BLOWUP_FACTOR )
		tcontext->cur_ccthresh = 2 * CWND_BLOWUP_FACTOR;

}

struct timing_algo tcp_reno = {
	.on_rcv = &reno_on_rcv,
	.on_drop = &reno_on_drop,
	.on_update_timeout = &reno_on_update_timeout,
	.on_late_response = &reno_on_late_response,
	.constructor = &reno_constructor,
	.destructor = &reno_destructor,
};
