/* 
        PortBunny kernel-module.
		
	Authors:        Fabian Yamaguchi <fabs@recurity-labs.com>

	Changes:        fabs        :       Initial Revision     28.03.07 
	
	Descripion:     Device-File kernel/userland-interface.
	
	
*/

#include "../scanner_module.h"
#include "scanner_ui.h"
#include "scanner_output_queue.h"
#include "pending_commands_queue.h"

#include <linux/fs.h>
#include <linux/list.h>
#include <asm/uaccess.h>


#define RECURITY_SCANNER_MAJOR 60
#define RECURITY_SCANNER_DEV_NAME "portbunny"

#define MAX_CMD_SIZE 32768

/**
   \addtogroup ScannerUI
@{
*/

/**
        scanner_ui_read:
	
	Called by kernel when a user-space-process reads from
	             /dev/portbunny.

	
	@param filp not used.
	@param buffer buffer supplied by the user-space program.
	@param buf_length The user-space buffer's length.
	@param offset not used.  
	

        Effect:  Fills 'buffer' with the next string contained in the
	scanner-output-queue. If buffer is not big enough to hold the
	string, return as much data as possible and save the rest of
	the string so that it can be returned next time a read is issued.


*/


static ssize_t scanner_ui_read(struct file *filp,
			       char *buffer, size_t buf_length,
			       loff_t *offset)
{
	static char *out_buf = NULL;
	int out_buf_length;
	unsigned int t;
	char *new_buf;
	char *dest = buffer;
	char *src;

	
	if(!out_buf){
		out_buf = scanner_output_queue_head();
		
		if(!out_buf)        /* Is queue emtpy? */
			return 0;
	}

	/* out_buf is now initialized. */

	src = out_buf;
	out_buf_length = strlen(out_buf) + 1;
	
	
	
	if(out_buf_length <= buf_length){
		/* 
		   user-supplied buffer is large enough for this line.
		   Copy until all data is in user-space.
		*/
		
		for(t = 0; t < out_buf_length - 1; t++)
			put_user(*(src++), dest++);
		
		
		kfree(out_buf);
		out_buf = NULL;
		return out_buf_length - 1;
	}

	/* user-supplied-buffer is too small for the complete line,
	   so fill the user-space-buffer and adjust out_buf accordingly
	   for the next read-call. */
	
	for(t = 0; t < buf_length; t++)
		put_user(*(src++), dest++);
	
	/* cut off the copied data */
	new_buf = kmalloc(sizeof(char) * (out_buf_length - buf_length), GFP_KERNEL);
	if(!new_buf)
		return -ENOMEM;
	
	memcpy(new_buf, out_buf + buf_length*sizeof(char), out_buf_length - buf_length);
	
	kfree(out_buf);
	out_buf = new_buf;
	return buf_length;
		
}


/**
        scanner_ui_write:
	Description: Called by kernel when a user-space-process writes to
	/dev/portbunny.
	
	
	@param filp not used.
	@param buffer buffer supplied by the user-space program.
	@param buf_length The user-space buffer's length.
	@param offset not used.  
	

        Copy the contents of 'buffer' into a kernel-space buffer
	and append it to a static kernel-buffer. If the static
	kernel-buffer then contains a full command, add the
	command to the pending-commands-queue and clear the
	buffer.
	
        Notes:   A command is terminated with a newline-character
	and has the form "command_name arg1 ... argn\n"
	

*/

static ssize_t scanner_ui_write(struct file *filp,
				const char __user *buffer, size_t length,
				loff_t *offset)
{	
	static char kcmd_buf[MAX_CMD_SIZE];
	static boolean initialized = FALSE;
		
	/*
	  Allocate length + 1 chars so that we can store
	  and extra terminating \0.
	 */

	char *kbuf = kmalloc(sizeof(char) * (length + 1) , GFP_KERNEL);

	if(!kbuf)
		return -ENOMEM;
	
	kbuf[length] = '\0';

	/* Do not allow empty writes */
	if(length == 0){
		printk("empty write received\n");
		kfree(kbuf);
		return -EFAULT;
	}

	if(!initialized){
		printk("Initializing kcmd_buf\n");
		kcmd_buf[0] = '\0';
		initialized = TRUE;
	}

		
	if (copy_from_user(kbuf, buffer, length) != 0){
		printk("Error copying from user-space\n");
		kfree(kbuf);
		return -EFAULT;
	}
	
	/* The commands are text-based. If there is
	 * a \0 somewhere in the command, we are clearly
	 * being attacked. So just don't bother executing
	 * any more code for this input. */
	
	if(strlen(kbuf) != length){
		printk("Zero-byte poisoning\n");
		kfree(kbuf);
		return -EFAULT;
	}	       	
	
		
	
	/* If the command does not follow the convention 
	   of being no bigger than MAX_CMD_SIZE, discard it.*/
	if(strlen(kcmd_buf) + length + 1 > MAX_CMD_SIZE){
		printk("ERROR: Command too large: discarding.\n");
		kcmd_buf[0] = '\0';
		kfree(kbuf);
		return -EFAULT;
	}
	
	
	/* append newly received data to kcmd_buf */
	strncat(kcmd_buf, kbuf, MAX_CMD_SIZE - strlen(kcmd_buf) - 1);	

	/* if the second-last-char is '\n', we have
	 * received one or more complete commands. */
	
	if(kcmd_buf[strlen(kcmd_buf) - 1] == '\n'){
		char *cmd_start = kcmd_buf;
		char *cmd_end   = NULL;
		
		while((cmd_end = strchr(cmd_start, '\n'))){			
			char *new_cmd;
			
			/* don't accept empty commands */
			if(cmd_end == cmd_start)
				break;
			
			new_cmd = kmalloc(sizeof(char) * (cmd_end - cmd_start + 2), GFP_KERNEL);
			
			memcpy(new_cmd, cmd_start, sizeof(char) * (cmd_end - cmd_start + 1));
			new_cmd[(cmd_end - cmd_start) + 1] = '\0';
		
			/* add the command to the pending-commands-queue */		

			pending_commands_queue_add(new_cmd);
			
			kfree(new_cmd);
			
			cmd_start = cmd_end + 1;
			
		}

		
		kcmd_buf[0] = '\0';
	}
	
	kfree(kbuf);
	return length;
}

static struct file_operations ui_fops = {
	.read  = scanner_ui_read,
	.write =  scanner_ui_write,
	.open  =  NULL,
	.release = NULL,
};

int scanner_ui_init()
{
	int ret;
	
	/* Registering device */
	ret = register_chrdev(RECURITY_SCANNER_MAJOR, RECURITY_SCANNER_DEV_NAME, &ui_fops);
	
	if (ret < 0)
		return ret;

	return SUCCESS;
}

/*
  Once this function has been called, no new commands from
  the device-file will be read.
*/

void scanner_ui_fini()
{
	
	unregister_chrdev(RECURITY_SCANNER_MAJOR, RECURITY_SCANNER_DEV_NAME);
	
}

/** @} */
