#include <linux/version.h>
#include <linux/tcp.h>
#include <linux/udp.h>
#include <linux/skbuff.h>

#include "parse_tree.h"
#include "packet_receiver.h"
#include "sniffed_packet_queue.h"

#define ERROR_WHILE_PARSING        1
#define FALLBACK_TREE_INDEX       10

/**
   root of the parse-tree in use.
   This pointer is valid after parse_tree_init has
   been called.
   
   A pointer to the tree can then be obtained by
   calling \ref get_parse_tree_root.
   
*/

struct parse_tree_node *new_parse_tree_node(node_func f,
					     struct parse_tree_node *parent,
					     __u16 index_in_parent, __u16 max_kids)
{
	struct parse_tree_node *this;
	
	/* Allocate space for new node */
	this = kmalloc(sizeof(struct parse_tree_node), GFP_KERNEL);
	if(!this) return NULL;

  
	if(max_kids != 0){
		/* max_kids = 0 <=> This is a leaf */
		this->children = kmalloc(sizeof(struct parse_tree_node *) * max_kids, GFP_KERNEL);
		
		if(!this->children)
			return NULL;

		memset(this->children, 0, sizeof(struct parse_tree_node *) * max_kids);
		
	}else{	
		this->children = NULL;
	}
	
	
	this->max_children = max_kids;
	this->func = f;
	
	
	/* Register with parent node */
	
	if(parent != NULL)
		(parent->children)[index_in_parent] = this;
	
	return this;
}

void free_parse_tree_node(struct parse_tree_node *this)
{
	kfree(this->children);
	kfree(this);
}


void walk_parse_tree(struct parse_tree_node *node, struct sk_buff *skb, void **p_descr)
{
	while(node->children != NULL){
		
		__u16 new_index =
			(node->func) (skb, node->max_children, p_descr);
		
		/* A packet that we do not handle has arrived */
		if(node->children[new_index] == NULL){			
			if(*p_descr)
				kfree(*p_descr);
			return;
		}		

		node = node->children[new_index];
	}

	/* We've reached a leaf */

	node->func(skb, 0, p_descr);
	
}

/**
   save the packet-description in
   the sniffed-packet-queue.
*/

int forward_to_packet_queue(void **p_descr)
{
	
	struct sniffed_packet_descr_t *descr = *p_descr;	
	struct sniffed_packet_queue_node *packet_queue =
		packet_receiver.sniffed_packet_queue;		

	/* Now add the packet-description structure to the
	   sniffed-packet-queue of trigger-state.
	*/
	sniffed_packet_queue_add(packet_queue, (struct sniffed_packet_descr_t *) descr);
	
	return 0;
}


/***************************************************
      Node-functions
****************************************************/


static __u16 initial_node(struct sk_buff *skb, __u16 max_children, void **p_descr)
{					
	__u16 retval = 0;
        					
	struct sniffed_packet_descr_t *descr;
	
	/* filter out any ethernet-frames which are not
	 * big enough to contain the ip-header.
	 */
	
	if(! pskb_may_pull(skb, sizeof(struct iphdr)))
		return ERROR_WHILE_PARSING;

	
	/* create empty packet-description and set
	 * time_received-field.
	 */

	descr = 
		*p_descr = new_sniffed_packet_descr();
	
	if(!(*p_descr))
		retval = ERROR_WHILE_PARSING;					
	
	return retval;
}

__u16 ip_protocol_mapper(struct sk_buff *skb, __u16 max_children, void **p_descr)
{

	/* In Kernel 2.6.22 sk_buff was changed
	*/
	
	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,22)
	struct iphdr *ipip_header = ipip_hdr(skb);	
	#else
	struct iphdr *ipip_header = skb->h.ipiph;
	#endif


	struct sniffed_packet_descr_t *descr;
	
	/* write source-address into packet-description-structure */
	descr = *p_descr;
	descr->subject = descr->src = ipip_header->saddr;
	
	/* Set the transport-layer pointer for following nodes.
	 * Check http://oss.sgi.com/archives/netdev/2002-02/msg00088.html
	 * for an explanation of why this must be done. */

	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,22)

	skb_set_transport_header(skb, ip_hdr(skb)->ihl *4);
	return ip_hdr(skb)->protocol % (max_children - 1);

	#else

	skb->h.raw = skb->nh.raw + skb->nh.iph->ihl*4;
	return skb->nh.iph->protocol % (max_children - 1);
	
	#endif

	
}

__u16 icmp_type_mapper(struct sk_buff *skb, __u16 max_children, void **p_descr )
{
	/* In Kernel 2.6.22 they changed sk_buff... */

	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,22)

	struct icmphdr *icmp_header = icmp_hdr(skb);
	
	#else
	
	struct icmphdr *icmp_header = skb->h.icmph;
	
	#endif
	

	struct sniffed_packet_descr_t *descr = *p_descr;
	descr->protocol = IPPROTO_ICMP;
	
	descr->icmp_seq_num = icmp_header->un.echo.sequence;
	descr->icmp_id = icmp_header->un.echo.id;
	
	return icmp_header->type % (max_children - 1);
}

__u16 tcp_handler(struct sk_buff *skb, __u16 m, void **p_descr)
{		
	/* In Kernel 2.6.22 they changed sk_buff... */

	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,22)

	struct tcphdr *tcp_header = tcp_hdr(skb);
	
	#else
	
	struct tcphdr *tcp_header = skb->h.th;
	
	#endif


	struct sniffed_packet_descr_t *descr = *p_descr;
	
	descr->protocol = IPPROTO_TCP;
	descr->port = ntohs(tcp_header->source);	
	descr->ack_seq = ntohl( tcp_header->ack_seq);
	
	if(tcp_header->syn && tcp_header->ack) return TCP_SYN_ACK_PACK;
	if(tcp_header->ack && tcp_header->rst) return TCP_ACK_RST_PACK;
	if(tcp_header->rst) return TCP_RST_PACK;

	return TCP_NOT_HANDLED;
}


static __u16 icmp_er_handler(struct sk_buff *skb, __u16 max_children, void **p_descr)
{
	
	struct sniffed_packet_descr_t *descr = *p_descr;
	descr->icmp_type = ICMP_ECHOREPLY;
	
	return forward_to_packet_queue(p_descr);
	
}

static __u16 icmp_tsr_handler(struct sk_buff *skb, __u16 max_children, void **p_descr)
{
	struct sniffed_packet_descr_t *descr = *p_descr;
	descr->icmp_type = ICMP_TIMESTAMPREPLY;

	return forward_to_packet_queue(p_descr);
}

static __u16 icmp_addr_mask_handler(struct sk_buff *skb, __u16 max_children, void **p_descr)
{
	struct sniffed_packet_descr_t *descr = *p_descr;
	descr->icmp_type = ICMP_ADDRESSREPLY;
	
	return forward_to_packet_queue(p_descr);
}

static __u16 icmp_dest_unreach_handler(struct sk_buff *skb, __u16 max_children, void **p_descr)
{
	struct sniffed_packet_descr_t *descr = *p_descr;
	descr->icmp_type = ICMP_DEST_UNREACH;
	
	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,22)

	descr->icmp_code = icmp_hdr(skb)->code;
	
	#else
	
	descr->icmp_code = skb->h.icmph->code;

	#endif

	
	/* The ICMP-dest-unreachables handled below include the packet which generated 
	   this response so we need to parse this original packet to understand
	   what this ICMP-dest-unreachable tells us.
	*/
	
	
	
	if(skb->len < (2 * sizeof(struct iphdr) + sizeof(struct icmphdr))){
		kfree(*p_descr);			
		return 0;
	}else{
		
		/* extract the packet which generated this response */
		struct iphdr *ipip_header = (struct iphdr *) (skb->data + sizeof(struct iphdr) + sizeof(struct icmphdr));	
		__u8 protocol = ipip_header->protocol;
		
		
		if(protocol == IPPROTO_UDP){
			struct udphdr *udp = (struct udphdr *) (skb->data + 2 * sizeof(struct iphdr)
								+ sizeof(struct icmphdr));		
			
			if(skb->len < (2 * sizeof(struct iphdr) + sizeof(struct icmphdr) + sizeof(struct udphdr))){
				kfree(*p_descr);
				return 0;
			}

		
			descr->subject = ipip_header->daddr;
			descr->port = ntohs(udp->dest);
			
			if(ntohs(udp->len) >= (sizeof(struct udphdr) + sizeof(__be16)) ){
				__be16 *data_addr = (__be16 *) udp;
				data_addr += (sizeof(struct udphdr) / sizeof(__be16));
				
				descr->udp_data = ntohs(*data_addr);
				
			}
	

		}else if(protocol == IPPROTO_TCP){
			
			struct tcphdr *tcp = (struct tcphdr *) (skb->data + 2 * sizeof(struct iphdr)
								+ sizeof(struct icmphdr));		

			
			/* check if packet is big enough to contain at least the first two fields of
			 * the TCP-header.
			 */

			if(skb->len < (2 * sizeof(struct iphdr) + sizeof(struct icmphdr) + 2 * sizeof(__be16))){
				kfree(*p_descr);
				printk("REACHED\n");
				return 0;
			}
			

			descr->subject = ipip_header->daddr;
			descr->port = ntohs(tcp->dest);								
	
		}
		
	}
	
	
	return forward_to_packet_queue(p_descr);
}


static __u16 tcp_syn_ack_handler(struct sk_buff *skb, __u16 m, void **p_descr)
{
	struct sniffed_packet_descr_t *descr = *p_descr;
	
	descr->tcp_syn_ack = TRUE;
	return forward_to_packet_queue(p_descr);
}

static __u16 tcp_ack_rst_handler(struct sk_buff *skb, __u16 m, void **p_descr)
{
	struct sniffed_packet_descr_t *descr = *p_descr;
	
	descr->tcp_ack_rst = TRUE;
	return forward_to_packet_queue(p_descr);	
}

static __u16 tcp_rst_handler(struct sk_buff *skb, __u16 m, void **p_descr)
{
	struct sniffed_packet_descr_t *descr = *p_descr;	
	
	descr->tcp_rst = TRUE;
	return forward_to_packet_queue(p_descr);	
}


static __u16 udp_handler(struct sk_buff *skb, __u16 max_children, void **p_descr)
{
	kfree(*p_descr);
	return 0;
}


struct parse_tree_node *new_parse_tree(void)
{
	
	struct parse_tree_node *root_node;
	struct parse_tree_node *ip_node;
	struct parse_tree_node *icmp_node;
	struct parse_tree_node *tcp_node;

	struct parse_tree_node *icmp_er_node;
	struct parse_tree_node *icmp_tsr_node;
	struct parse_tree_node *icmp_addr_mask_node;
	struct parse_tree_node *icmp_dest_unreach_node;
	
	struct parse_tree_node *tcp_syn_ack_node;
	struct parse_tree_node *tcp_ack_rst_node;
	struct parse_tree_node *tcp_rst_node;
	struct parse_tree_node *udp_node;
	
	root_node = 
		new_parse_tree_node(initial_node, NULL, 0, 2);
	
	if(!root_node) return NULL;

	root_node->children[ERROR_WHILE_PARSING] = NULL;
	
	ip_node = new_parse_tree_node(&ip_protocol_mapper,
				       root_node, 0, MAX_INET_PROTOS);
	
	
	if(!ip_node) return NULL;
	
	/* icmp-type-node */
	
	icmp_node = new_parse_tree_node(&icmp_type_mapper,
					 ip_node, IPPROTO_ICMP, NR_ICMP_TYPES); 

	if(!icmp_node) return NULL;

	/* tcp-node */

	tcp_node = new_parse_tree_node(&tcp_handler,
					ip_node, IPPROTO_TCP, N_TCP_TYPES);
	if(!tcp_node) return NULL;
	

	icmp_er_node = new_parse_tree_node(&icmp_er_handler,
					    icmp_node, ICMP_ECHOREPLY, 0);
	
	if(!icmp_er_node) return NULL;
	
	icmp_tsr_node = new_parse_tree_node(&icmp_tsr_handler,
					     icmp_node, ICMP_TIMESTAMPREPLY, 0);
	
	if(!icmp_tsr_node) return NULL;

	icmp_addr_mask_node = new_parse_tree_node(&icmp_addr_mask_handler,
						   icmp_node, ICMP_ADDRESSREPLY, 0);
	
	if(!icmp_addr_mask_node) return NULL;
	

	icmp_dest_unreach_node = new_parse_tree_node(&icmp_dest_unreach_handler,
						      icmp_node, ICMP_DEST_UNREACH, 0);
	
	if(!icmp_dest_unreach_node) return NULL;
	

	tcp_syn_ack_node = new_parse_tree_node(&tcp_syn_ack_handler,
						tcp_node, TCP_SYN_ACK_PACK, 0);
	if(!tcp_syn_ack_node) return NULL;
	
	tcp_ack_rst_node = new_parse_tree_node(&tcp_ack_rst_handler,
						tcp_node, TCP_ACK_RST_PACK, 0);
	
	if(!tcp_ack_rst_node) return NULL;
	
	tcp_rst_node = new_parse_tree_node(&tcp_rst_handler,
					   tcp_node, TCP_RST_PACK, 0
					   );

	if(!tcp_rst_node) return NULL;

	/* udp-node */
	
	udp_node = new_parse_tree_node(&udp_handler,
					ip_node, IPPROTO_UDP, 0);
	
	if(!udp_node) return NULL;
	

	return root_node;

}

void free_parse_tree(struct parse_tree_node *root)
{
	int t;
	
	if(root == NULL) return;

	/* free all sub-trees */

	for(t = 0; t < root->max_children; t++)
		if(root->children[t])
			free_parse_tree(root->children[t]);
	
	
	free_parse_tree_node(root);	
}


