#include "sniffed_packet_queue.h"

#include <linux/module.h>
#include <linux/interrupt.h>

/*
  
  This function is used:
  
  (1) By packet-receivers to store packet-descriptions in
  the classified-packet-queues of a state.
  
  (2) By the scan-job-manager to store packet-descriptions
  in classified-packet-queues of single scan-jobs.

    
  Arguments:
  
  this:        The classified-packet-queue to store in.
  
  p_descr:     A pointer to the packet-description structure
               that is to be stored.

  Attention:
  
  The contents of p_descr is NOT copied so the caller may
  not free p_descr after calling this function.
  

  Requirements:
  
  (1) p_descr is fully initialized.
  
*/

void sniffed_packet_queue_add(struct sniffed_packet_queue_node *this,
				 struct sniffed_packet_descr_t *p_descr)
{
	struct sniffed_packet_queue_node *new_node =
		kmalloc(sizeof(struct sniffed_packet_queue_node), GFP_ATOMIC);
	
	if(!new_node) return;
	
	write_lock(this->lock);

	new_node->packet_descr = p_descr;
	INIT_LIST_HEAD(&new_node->list);
	
	list_add_tail(&new_node->list, &this->list);
	
	write_unlock(this->lock);
	
}

/*
        This function is called by the scan-job-manager
	(by function 'process_pending_commands') to
	retrieve packet-descriptions from classified-
	packet-queues of a state and pass them on to
	classified-packet-queues of single scan-jobs.

        Requirements:
	called from user-context.
*/

struct sniffed_packet_descr_t *
sniffed_packet_queue_head(struct sniffed_packet_queue_node *this)
{
	struct sniffed_packet_queue_node *entry;
	struct sniffed_packet_descr_t *retval;
	
	write_lock_bh(this->lock);

	if(list_empty(&this->list)){
		write_unlock_bh(this->lock);
		return NULL;
	}

	entry = list_entry( this->list.next,
			    struct sniffed_packet_queue_node, list);

	retval = entry->packet_descr;

	list_del(&entry->list);
	kfree(entry);


	write_unlock_bh(this->lock);
	return retval;
}

