
#include "cmd_handlers.h"
#include "trigger_finding_methods.h"
#include "state.h"

#include "../scanner_ui/cmd_handlers.h"
#include "../scanner_ui/scanner_output_queue.h"

#include <linux/list.h>
#include <linux/module.h>

#include "icmp_er_method.h"
#include "tcp_syn_method.h"
#include "udp_method.h"


/**
   Command: "append_to_methods_list $TARGET_IP TRIGGER $METHOD_NAME"
   
   no response.

   Adds the trigger-method $METHOD_NAME to the list of triggers tried.

*/

void append_to_methods_list_handler(struct command_t *cmd, struct scan_job_t *scan_job)
{
	struct trigger_state_context *context =
		(struct trigger_state_context *) scan_job->state_context;
        int method_id; 
	struct trigger_instance *node;
	
	enum {
		TARGET_IP = 0,
		STATE,
		METHOD_NAME,
		NARGS
	};

	if(scan_job->active) return;
	
	if(cmd->argc != NARGS){
		scanner_output_queue_add("ERROR -1 Wrong number of parameters\n");
		return;
	}
	
	method_id = method_id_by_method_name(cmd->argv[METHOD_NAME]);

	if(method_id == NO_FINDING_METHOD)
		return;
	
	node = new_trigger_instance(method_id, ALL_ROUNDS);

	if(!node){
		scanner_output_queue_add("ERROR -1 No memory left\n");
		return;
	}	
	
	list_add_tail(&node->list, &context->trigger_methods_list.list);
	
	
}

/**
   Command: "clear_methods_list $TARGET_IP TRIGGER"
   
   no response.

   clears the list of triggers used.
   
*/


void clear_methods_list_handler(struct command_t *cmd, struct scan_job_t *scan_job)
{
	struct trigger_state_context *context =
		(struct trigger_state_context *) scan_job->state_context;

	struct trigger_instance *root = 
		&(context->trigger_methods_list);

	if(scan_job->active) return;

	trigger_finding_method_list_clear(root);
}

/**
   Command: "number_of_triggers_wanted $TARGET_IP TRIGGER $NTRIGGERS_WANTED"
   
   on success: no response.

   Sets the number of triggers wanted. Once this number of triggers has been
   identified, the trigger-state-scanjob will be finished.
   
*/

void number_of_triggers_wanted_handler(struct command_t *cmd, struct scan_job_t *scan_job)
{
	struct trigger_state_context *context =
		(struct trigger_state_context *) scan_job->state_context;
	char **endp = NULL;

	enum {
		TARGET_IP = 0,
		STATE,
		NTRIGGERS_WANTED,
		NARGS
	};

	if(scan_job->active) return;

	if(cmd->argc != NARGS){
		scanner_output_queue_add("ERROR -1 Wrong number of parameters\n");
		return;
	}
	
	context->ntriggers_to_find = simple_strtoul(cmd->argv[NTRIGGERS_WANTED], endp, 10);
}

/**
   Command: "set_method_timeout $TARGET_IP $SECONDS $NSECONDS"
   
   on success: no response.

   Sets the timeout to $SECONDS seconds + $NSECONDS nano-seconds.
   
*/

void set_method_timeout_handler(struct command_t *cmd, struct scan_job_t *scan_job)
{
	struct trigger_state_context *context =
		(struct trigger_state_context *) scan_job->state_context;
	char **endp = NULL;

	enum {
		TARGET_IP = 0,
		SECONDS,
		NSECONDS,
		NARGS
	};

	if(scan_job->active) return;
		
	
	if(cmd->argc != NARGS){
		scanner_output_queue_add("ERROR -1 Wrong number of parameters\n");
		return;
	}
	
	
	context->timeout.tv_sec = simple_strtoul(cmd->argv[SECONDS], endp, 10);
	context->timeout.tv_nsec = simple_strtoul(cmd->argv[NSECONDS], endp, 10);
	
}

/**
   Command: "set_tcp_syn_ports $TARGET_IP TRIGGER $PORT1 ... $PORTn"
   
   on success: no response.

   Sets the ports that the tcp-syn method will try to $PORT1... $PORTn.
   
*/

void set_tcp_syn_ports_handler(struct command_t *cmd, struct scan_job_t *scan_job)
{
	/* argv[2] ... argv[n]: ports to use */
	
	/* cmd->argc is at least two because otherwise the scan-job could have
	   not been determined.
	*/
	
	struct trigger_state_context *context =
		(struct trigger_state_context *) scan_job->state_context;
	
	struct trigger_instance *node;
	struct tcp_syn_context *syn_context;		
	
	char **endp = NULL;

	int new_nports;
	int *new_ports;
	
	int t;
	
	if(scan_job->active) return;

	/* nothing to be done if no ports were specified. */
	if(cmd->argc < 3) return;

	node = get_trigger_node_by_method_id(&context->trigger_methods_list,
					     TCP_SYN_FINDING_METHOD);
	
	if(!node){
		scanner_output_queue_add("ERROR -1 TCP_SYN-trigger currently not used\n");
		return;
	}
	
	syn_context = node->context;
	
	new_nports = cmd->argc - 2;
	new_ports = kmalloc(sizeof(int) * new_nports, GFP_KERNEL);

	if(!new_ports){
		scanner_output_queue_add("ERROR -1 Out of memory\n");
		return;
	}
	
	/* Copy ports into newly allocated array */
	for(t = 0; t < new_nports; t++){
		new_ports[t] = simple_strtoul(cmd->argv[t+2], endp, 10);
	}
	

	kfree(syn_context->ports);
	syn_context->nports = new_nports;
	syn_context->ports = new_ports;

	
}


/**
   Command: "set_udp_ports $TARGET_IP TRIGGER $PORT1 ... $PORTn"
   
   on success: no response.

   Sets the ports that the udp-method will try to $PORT1... $PORTn.
   
*/

void set_udp_ports_handler(struct command_t *cmd, struct scan_job_t *scan_job)
{
	/* argv[2] ... argv[n]: ports to use */
	
	/* cmd->argc is at least two because otherwise the scan-job could have
	   not been determined.
	*/
	
	struct trigger_state_context *context =
		(struct trigger_state_context *) scan_job->state_context;
	
	struct trigger_instance *node;
	struct udp_context *udp_context;
	
	char **endp = NULL;

	int new_nports;
	int *new_ports;
	int t;
	
	if(scan_job->active) return;

	/* nothing to be done if no ports were specified. */
	if(cmd->argc < 3) return;

	node = get_trigger_node_by_method_id(&context->trigger_methods_list,
					     UDP_FINDING_METHOD);
	
	if(!node){
		scanner_output_queue_add("ERROR -1 UDP-trigger currently not used\n");
		return;
	}
	
	udp_context = node->context;
	

	new_nports = cmd->argc - 2;
	new_ports = kmalloc(sizeof(int) * new_nports, GFP_KERNEL);

	if(!new_ports){
		scanner_output_queue_add("ERROR -1 Out of memory\n");
		return;
	}
	
	/* Copy ports into newly allocated array */
	for(t = 0; t < new_nports; t++){
		new_ports[t] = simple_strtoul(cmd->argv[t+2], endp, 10);
	}
	
	
	kfree(udp_context->ports);
	udp_context->nports = new_nports;
	udp_context->ports = new_ports;
}

