#include "icmp_er_method.h"
#include "trigger_finding_methods.h"
#include "state.h"

#include "../scanner_module.h"
#include "../packet_submitter.h"
#include "../scan_job_states.h"

#include "../flood_state/packet_batch.h"

#include <linux/time.h>
#include <net/ip.h>


static int icmp_er_receiver(struct scan_job_t *this,
			    struct trigger_instance *method_instance)
{
	
	struct sniffed_packet_descr_t *p_descr;
	struct trigger_state_context *context =
		(struct trigger_state_context *) this->state_context;
	
	int reported = 0;
	

	while((p_descr = sniffed_packet_queue_head(&this->packets))){
		struct sniffed_packet_descr_t *descr =
			(struct sniffed_packet_descr_t *) p_descr;
		
		/* If protocol and icmp-type is correct, the packet
		   is from the host this scan-job is scanning and
		   this trigger has not been reported, report it.
		*/
		if( (descr->protocol == IPPROTO_ICMP) &&
		    (descr->icmp_type == ICMP_ECHOREPLY) &&
		    (!reported) && (descr->src == this->addr)){

			trigger_instance_out(method_instance, this, 0);
			context->ntriggers_found++;
			reported = 1;			
		}			


		kfree(p_descr);
	}
	
	if(!reported)
		return CALL_AGAIN;
	else
		return FINISHED;
	
}


/*
  Options: icmp_er_data_size
           icmp_er_data

  Note: Currently the trigger-sending-functions
  are used by flood-state as well as trigger-state
  scanjob's which means you'll have to be careful
  when dereferencing this->state_context. Depending
  on this->state it can either be a trigger_state_context
  or a flood_state_context.
  
	   
*/

static void icmp_er_sender(struct scan_job_t *this,
			   struct trigger_instance *method_instance)
{
	struct icmp_er_context *context = method_instance->context;
	int icmp_data_size = context->data_len;
	char *icmp_data = kmalloc(icmp_data_size, GFP_KERNEL);	
	
	
	if(!icmp_data) return;
	memcpy(icmp_data, context->data, context->data_len);
	
	
	send_icmp_packet(this->addr, ICMP_ECHO,
			 0, icmp_data_size,
			 icmp_data, context->batch_id);
	
}


static int icmp_er_context_init(struct trigger_instance *this,
				int unused)
{
	
	const int default_data_len = 1;
	const char default_data_char = 'a';

	/* Allocate space for icmp_er_context */
	struct icmp_er_context *er_context = 
		this->context = kmalloc(sizeof(struct icmp_er_context), GFP_KERNEL);
	
	if(!er_context)
		return FAILURE;
	
	/* register some sane default-values */
	er_context->data_len = default_data_len;
	er_context->data = kmalloc(sizeof(char) * default_data_len, GFP_KERNEL);
	
	if(!er_context->data){
		kfree(this->context);
		return FAILURE;
	}
	
	memset(er_context->data, default_data_char, default_data_len);
	
	
	return SUCCESS;
}

static int icmp_er_is_response(struct sniffed_packet_descr_t *descr,
			       struct packet_batch *batch)
{
	
	unsigned int batch_id = descr->icmp_id;
	

	return (batch_id == batch->port_indices[0]) &&
		(descr->protocol == IPPROTO_ICMP) &&
		(descr->icmp_type == ICMP_ECHOREPLY);
	
}

static int icmp_er_extract_batch_id(struct sniffed_packet_descr_t *descr)
{
	if ((descr->protocol == IPPROTO_ICMP) &&
	    (descr->icmp_type == ICMP_ECHOREPLY))
		return descr->icmp_id;
	
	return -1;
}


static void icmp_register_batch_id(struct trigger_instance *this,
				   unsigned int batch_id)
{
	struct icmp_er_context *er_context = this->context;
		

	/* Save batch_id in id */
	er_context->batch_id = batch_id;
       	
	

}


static void icmp_er_context_fini(struct trigger_instance *this)
{
	struct icmp_er_context *er_context = this->context;	

	kfree(er_context->data);
	kfree(er_context);
	
	
}

static void *
icmp_er_context_copy(struct trigger_instance *this)
{
	struct icmp_er_context *context = this->context;
	
	struct icmp_er_context *new_context = 
		kmalloc(sizeof(struct icmp_er_context), GFP_KERNEL);

	if(!new_context)
		return NULL;

	memcpy(new_context, context, sizeof(struct icmp_er_context));
	
	/* Copy data-field */

	new_context->data = kmalloc(sizeof(char) * context->data_len, GFP_KERNEL);

	if(!new_context->data){
		kfree(new_context);
		return NULL;
	}

	memcpy(new_context->data, context->data, sizeof(char) * context->data_len);
	
	return new_context;
	
}

static int icmp_er_get_round_by_descr(struct sniffed_packet_descr_t *descr)
{
	return 0;
}

static int icmp_er_get_round(void *context)
{
	return 0;
}

struct trigger_finding_method icmp_er_method = {
	
	.name = "ICMP_ER",
	.sender_func = &icmp_er_sender,
	.receiver_func = &icmp_er_receiver,
	
	.is_response = &icmp_er_is_response,
	.register_batch_id = &icmp_register_batch_id,
	.extract_batch_id  = &icmp_er_extract_batch_id,

	.get_round_by_descr = &icmp_er_get_round_by_descr,
	.get_round          = &icmp_er_get_round,

	.context_init = icmp_er_context_init,
	.context_fini = icmp_er_context_fini,
	
	.context_copy = icmp_er_context_copy,		
	
};
