
#include "../scanner_module.h"
#include "../scan_job_states.h"
#include "../sniffer/sniffed_packet_queue.h"
#include "../scanner_ui/cmd_handlers.h"
#include "../queue.h"
#include "../packet_submitter.h"
#include "../timespec_utils.h"
#include "state.h"
#include "trigger_finding_methods.h"
#include "cmd_handlers.h"

#include <linux/list.h>


/** \addtogroup TriggerStateScanJobManager
    @{
*/

static int has_timed_out(struct scan_job_t *this);

static struct cmd_handlers_hash_bucket cmd_handlers_hash[CMD_HASH_SIZE];

static int trigger_state_init(void)
{	
	cmd_handlers_init(trigger_state.cmd_handlers);
	
	/* now register all cmd-handlers */
	
	cmd_handlers_register(trigger_state.cmd_handlers,
			      "append_to_methods_list",
			      &append_to_methods_list_handler);
	
	cmd_handlers_register(trigger_state.cmd_handlers,
			      "clear_methods_list",
			      &clear_methods_list_handler);
	
	cmd_handlers_register(trigger_state.cmd_handlers,
			      "number_of_triggers_wanted",
			      &number_of_triggers_wanted_handler);
	
	cmd_handlers_register(trigger_state.cmd_handlers,
			      "set_method_timeout",
			      &set_method_timeout_handler);	
	
		
	cmd_handlers_register(trigger_state.cmd_handlers,
			      "set_tcp_syn_ports",
			      &set_tcp_syn_ports_handler);	
	
	cmd_handlers_register(trigger_state.cmd_handlers,
			      "set_udp_ports",
			      &set_udp_ports_handler);
			
	return SUCCESS;
}


static void trigger_state_fini(void)
{
	cmd_handlers_fini(trigger_state.cmd_handlers);
}

/**
   TriggerStateScanJob-Constructor  
   
   Constructs a trigger-state scan-job's context.
   
   Return-value: SUCCESS on success, FAILURE otherwise.
*/

static int scan_job_init(struct scan_job_t *this)
{
	struct trigger_state_context *context;
	
	printk("Constructing a trigger-state scan-job using trigger_state/scan_job_init\n");

        /* allocate space for context */
	context = this->state_context
		= kmalloc(sizeof(struct trigger_state_context), GFP_KERNEL);
	
	if(!this->state_context){
		this->state_context = NULL;
		return FAILURE;
	}

	/* Initialize context */
	memset(this->state_context, 0, sizeof(struct trigger_state_context));
	INIT_LIST_HEAD(&context->trigger_methods_list.list);
	
	context->last_rounds_finding_method = NO_FINDING_METHOD;
	context->ntriggers_found = 0;
	
	/* To speed up the trigger process, try to get only
	 * 1 trigger by default */
	context->ntriggers_to_find = 1;
	
	/* initialize the packet-queue used for incoming packets */
	/* if the number of state-independent members which need
	   initialization grows, create a generic scan-job-constructor. */


	INIT_LIST_HEAD(&this->packets.list);
	this->packets.lock = kmalloc(sizeof(rwlock_t), GFP_KERNEL);
	
	if(!this->packets.lock){
		kfree(this->state_context);
		this->state_context = NULL;
		return FAILURE;
	}

	*this->packets.lock = RW_LOCK_UNLOCKED;
	
	context->timeout.tv_sec = TRIGGER_DEFAULT_TIMEOUT_S;
	context->timeout.tv_nsec = TRIGGER_DEFAULT_TIMEOUT_NS;	
	
	return trigger_finding_method_list_init(this);
}

/**
   TriggerStateScanJob-Destructor
   
*/

static void scan_job_fini(struct scan_job_t *this)
{
	struct trigger_state_context *context
		= this->state_context;
	
	trigger_finding_method_list_clear( &context->trigger_methods_list );
		
	if(this->packets.lock)
		kfree(this->packets.lock);
	
	if(this->addr_str)
		kfree(this->addr_str);
	
	/* free the scan-job */
	kfree(this->state_context);
	kfree(this);
}


/**
   Scan-job-manager-function:
   
   Return-Value: Returns FINISHED if it has finished and CALL_AGAIN if
   it has not.
*/

static int trigger_state_scan_job_manager(struct scan_job_t *this)
{
	struct trigger_state_context *context = this->state_context;
	int new_method_id;
	struct timespec timeout;
	int finished;	

	
	if(context->last_rounds_finding_method != NO_FINDING_METHOD){
		
		
		/* 
		   we sent a trigger last round,
		   let's call the receiver-func for this method
		   to check for answers
		*/
		
		finished = trigger_finding_methods[context->
						     last_rounds_finding_method]->receiver_func(this, context->cur_finding_node);
		
		
		/* if receiver is not finished and has not
		   timed out, call us again. */
		if( (!finished) && !has_timed_out(this) )
			return CALL_AGAIN;
		
		
		context->last_rounds_finding_method = NO_FINDING_METHOD;		
		del_trigger_instance(context->cur_finding_node);
	}

	
	
	/* Return FINISHED if we have found enough triggers */
	if(context->ntriggers_to_find > 0){
		if(context->ntriggers_found >= context->ntriggers_to_find){
			printk("Found %d triggers, that's enough :)\n", context->ntriggers_found);
			return FINISHED;
		}
	}
	
	/* next trigger, please.*/
	
	context->cur_finding_node = trigger_finding_method_list_head(this);
	
	if(context->cur_finding_node == NULL)
		/* no triggers left, we're done */
		return FINISHED;
	
	
	new_method_id = context->cur_finding_node->method_id;
	timeout = context->timeout;
	
	
	/*
	  Now launch this trigger-finding-method's sender-routine.
	*/
	
	trigger_finding_methods[new_method_id]->sender_func(this, context->cur_finding_node);
	
	/* set timeout-time to be current-time + timeout-value
	   specified by the trigger-method we are currently using.
	 */
	
	context->timeout_spec = CURRENT_TIME;
	
	context->timeout_spec.tv_sec += timeout.tv_sec;
	timespec_add_ns(&context->timeout_spec, timeout.tv_nsec);
       	
	
	/* write down current trigger-method-id for next call. */
	context->last_rounds_finding_method = new_method_id;
	
	return CALL_AGAIN;
	
}


static int has_timed_out(struct scan_job_t *this)
{
	struct trigger_state_context *context = this->state_context;

	/* retrieve current time */
	struct timespec cur_time = CURRENT_TIME;
	
	/* has time-out been reached? */

	return (timespec_compare(&context->timeout_spec, &cur_time) <= 0);	
}

struct scan_job_state_t trigger_state = {
	
	
	.name = "TRIGGER",
	.scan_job_manager = &trigger_state_scan_job_manager,
	.scan_job_init = &scan_job_init,
	.scan_job_fini = &scan_job_fini,
	.init = &trigger_state_init,
	.fini = &trigger_state_fini,

	.cmd_handlers = cmd_handlers_hash,

};

/** @} */
