#include "state.h"
#include "tcp_syn_method.h"

#include "trigger_finding_methods.h"

#include "../scanner_module.h"
#include "../packet_submitter.h"
#include "../scan_job_states.h"

#include "../flood_state/packet_batch.h"

#include <linux/time.h>
#include <net/ip.h>



int DEFAULT_NUM_PORTS_TCP =  5;
int DEFAULT_PORTS_TCP[] = {80, 443, 22, 445, 113};

#define TRIGGER_SRC_PORT 61373
#define MAX_TRIGGER_SRC_PORT 62390

static int tcp_syn_context_init(struct trigger_instance *this, int round)
{
	
	/*
	  Allocate memory for context and copy default
	  port-list into context.
	*/
	
	struct tcp_syn_context *syn_context =
		this->context = kmalloc(sizeof(struct tcp_syn_context), GFP_KERNEL);

	if(!syn_context)
		return FAILURE;
	
	if(round != ALL_ROUNDS)
		syn_context->nports = 1;
	else
		syn_context->nports = DEFAULT_NUM_PORTS_TCP;

	
	syn_context->ports = kmalloc(sizeof(int) * syn_context->nports, GFP_KERNEL);					

	if( (!syn_context->ports)){
		kfree(syn_context);
		return FAILURE;
	}

	
	if(round != ALL_ROUNDS)
		syn_context->ports[0] = round;
	else
		memcpy(syn_context->ports, DEFAULT_PORTS_TCP, sizeof(int) * syn_context->nports);
	
	
	syn_context->src_port = TRIGGER_SRC_PORT;
	
	return SUCCESS;
}


static int tcp_is_response(struct sniffed_packet_descr_t *descr,
			   struct packet_batch *batch)
{		
	struct trigger_instance *method_instance = 
		batch->trigger_node;
	struct tcp_syn_context *syn_context = 
		method_instance->context;
	int port;
	if( (!syn_context->ports) || (syn_context->nports != 1)){
		printk("INVALID method_instance!\n");
		return FALSE;
	}
		
	printk("descr->ack_seq / port_indices[0] %d/%d\n", descr->ack_seq, batch->port_indices[0]);

	port = syn_context->ports[0];	

	return (descr->protocol == IPPROTO_TCP) &&
		( (port == 0) || (descr->port == port) ) &&
		(((descr->ack_seq - 1) == batch->port_indices[0]));
}


static int tcp_syn_extract_batch_id(struct sniffed_packet_descr_t *descr)
{
	if ((descr->protocol == IPPROTO_TCP))
		/* filter batch-0 responses. */
		if(descr->ack_seq != 1)
			return (descr->ack_seq - 1);
		
	return -1;
}


/*
  
  Adds tcp-syn-packets for each port specified in
  the method-instance.
  
*/

static void tcp_syn_sender(struct scan_job_t *this,
			   struct trigger_instance *method_instance)
{
		
	struct trigger_state_context *context =
		this->state_context;
	struct tcp_syn_context *syn_context = 
		method_instance->context;
	
	int t;
	
	
	/* for each port in the port-list,
	 * create an outgoing-packet-descr and add it
	 * to the outgoing-packet-queue.
	 */
	
	for(t = 0; t < syn_context->nports; t++)		
		send_tcp_syn_packet(this->addr, syn_context->ports[t],
				    syn_context->src_port, syn_context->batch_id);
	

	/* write down the number of triggers we sent
	   so that the syn-receiver knows what to expect.
	   Also, reset nprobes_received.
	*/
	
	if(this->state == SCAN_JOB_STATE_TRIGGER_STATE){
	
		context->nprobes_received = 0;
		context->nprobes_sent = syn_context->nports;	
				
	}
	
}

static int get_round_by_port(struct trigger_instance *method_instance,
			     int port)
{
	struct tcp_syn_context *syn_context =
		method_instance->context;
	
	int t;
	for(t = 0; t < syn_context->nports; t++)
		if(syn_context->ports[t] == port) return t;

	return -1;

}


static int tcp_syn_receiver(struct scan_job_t *this,
			    struct trigger_instance *method_instance)
{
	struct sniffed_packet_descr_t *p_descr;
	struct trigger_state_context *context =
		this->state_context;
		

	while((p_descr = sniffed_packet_queue_head(&this->packets))){
		struct sniffed_packet_descr_t *descr =
			(struct sniffed_packet_descr_t *) p_descr;
					
		if( (descr->protocol == IPPROTO_TCP) &&  
		    (descr->src == this->addr)){
			
			int round = get_round_by_port(method_instance, descr->port);
			

			/* if a probe was actually sent on this port */
			if(round != -1){
											
			  
				trigger_instance_out(method_instance, this, descr->port);
				context->nprobes_received++;
				context->ntriggers_found++;
			}
		}			


		kfree(p_descr);
	}
	
	
	/* 
	   If answers for all probes have been received,
	   we are finished and must not wait for the timeout.
	*/
	
	if(context->nprobes_received == context->nprobes_sent)
		return FINISHED;
	else
		return CALL_AGAIN;
	
}

static void tcp_syn_register_batch_id(struct trigger_instance *this,
				      unsigned int batch_id)
{
	struct tcp_syn_context *context = this->context;

	context->batch_id = batch_id;
}


static void tcp_syn_context_fini(struct trigger_instance *this)
{
	struct tcp_syn_context *syn_context = this->context;
	
	if(syn_context->ports)
		kfree(syn_context->ports);
		
	kfree(syn_context);
}

static void *
tcp_syn_context_copy(struct trigger_instance *this)
{
	struct tcp_syn_context *context = this->context;
	
	struct tcp_syn_context *new_context = 
		kmalloc(sizeof(struct tcp_syn_context), GFP_KERNEL);

	if(!new_context)
		return NULL;

	memcpy(new_context, context, sizeof(struct tcp_syn_context));
	
	new_context->ports = kmalloc(sizeof(int) * context->nports, GFP_KERNEL);

	if(!new_context->ports){
		kfree(new_context);
		return NULL;
	}
			
	if(++context->src_port > MAX_TRIGGER_SRC_PORT){
		context->src_port = TRIGGER_SRC_PORT;
	}	
	
	memcpy(new_context->ports, context->ports, sizeof(int) * context->nports);
			
	return new_context;
	
}

static int tcp_syn_get_round_by_descr(struct sniffed_packet_descr_t *descr)
{
	return descr->port;
}

static int tcp_syn_get_round(void *context)
{
	struct tcp_syn_context *c = context;

	return c->ports[0];
}

struct trigger_finding_method tcp_syn_method = {
	.name = "TCP_SYN",
	
	.sender_func = &tcp_syn_sender,
	.receiver_func = &tcp_syn_receiver,

	.is_response = &tcp_is_response,
	
	.register_batch_id = &tcp_syn_register_batch_id,
	.extract_batch_id  = &tcp_syn_extract_batch_id,

	.get_round_by_descr = &tcp_syn_get_round_by_descr,
	.get_round          = &tcp_syn_get_round,

	.context_init = tcp_syn_context_init,
	.context_fini = tcp_syn_context_fini,

	.context_copy = tcp_syn_context_copy,
	
};


