#include "state.h"
#include "udp_method.h"
#include "trigger_finding_methods.h"

#include "../scanner_module.h"
#include "../packet_submitter.h"
#include "../scan_job_states.h"

#include "../flood_state/packet_batch.h"

#include <linux/time.h>
#include <net/ip.h>

int DEFAULT_NUM_PORTS_UDP = 2;
int DEFAULT_PORTS_UDP[] = {161, 162};
int DEFAULT_UDP_DELAY_NS = 500000000;
int DEFAULT_UDP_DELAY_S = 0;

static int udp_context_init(struct trigger_instance *this,
			    int round)
{
	
	/*
	  Allocate memory for context and copy default
	  port-list into context.
	*/
	
	struct udp_context *udp_context =
		this->context = kmalloc(sizeof(struct udp_context), GFP_KERNEL);

	if(!udp_context)
		return FAILURE;
		
	udp_context->delay.tv_sec = DEFAULT_UDP_DELAY_S;
	udp_context->delay.tv_nsec = DEFAULT_UDP_DELAY_NS;
	
	/* 
	 * Initialize a data-field of 32 bit which will be used 
	 * to store the packet_batch-id of the packet_batch
	 * the trigger was sent in. Only 16 bit are used.
	 */
	
	udp_context->data = 0;
	
	
	/* if a round was specified (round != ALL_ROUNDS),
	   create a port-list which contains the single
	   port specified by round.
	   Else, use default port-list.		 
	*/

	
	if(round != ALL_ROUNDS)
		udp_context->nports = 1;		
	else		
		udp_context->nports = DEFAULT_NUM_PORTS_UDP;	
	

	/* Allocate memory for port-list */
	udp_context->ports = kmalloc(sizeof(int) * udp_context->nports, GFP_KERNEL);		
		
	
	if(!udp_context->ports){
		kfree(udp_context);
		return FAILURE;
	}
	
	/* copy ports into newly allocated port-list */

	if(round != ALL_ROUNDS)
		udp_context->ports[0] = round;
	else
		memcpy(udp_context->ports, DEFAULT_PORTS_UDP, sizeof(int) * udp_context->nports);
	
	
	return SUCCESS;
}



static void udp_sender(struct scan_job_t *this,
		       struct trigger_instance *method_instance)
{
	
	int t;
	
	struct trigger_state_context *context =
		this->state_context;

	struct udp_context *udp_context = 
		method_instance->context;	
			

	/* for each port we wish to check,
	 * create an outgoing-packet-descr and add it
	 * to the outgoing-packet-queue.
	 */
	

	for(t = 0; t < udp_context->nports; t++)
		
		send_udp_packet(this->addr, udp_context->ports[t],
				udp_context->data);
	
	
	/* write down the number of triggers we sent
	   so that the syn-receiver knows what to expect.
	   Also reset nprobes_received.
	*/
	
	if(this->state == SCAN_JOB_STATE_TRIGGER_STATE){		
		  					
		context->nprobes_received = 0;
		context->nprobes_sent = udp_context->nports;						
		
	}
}

static int udp_receiver(struct scan_job_t *this,
			struct trigger_instance *method_instance)
{
	struct sniffed_packet_descr_t *p_descr;
	struct trigger_state_context *context =
		this->state_context;
	
	
	while((p_descr = sniffed_packet_queue_head(&this->packets))){
		struct sniffed_packet_descr_t *descr =
			(struct sniffed_packet_descr_t *) p_descr;
		
		/* If protocol and icmp-type is correct, the packet
		   is from the host this scan-job is scanning and
		   this trigger has not been reported, report it.
		*/
		if( (descr->protocol == IPPROTO_ICMP) &&
		    (descr->icmp_type == ICMP_DEST_UNREACH) &&
		    (descr->icmp_code == ICMP_PORT_UNREACH) &&
		    (descr->src == this->addr)){
				
			trigger_instance_out(method_instance, this, descr->port);
			context->nprobes_received++;
			context->ntriggers_found++;

		}			

		
		kfree(p_descr);
	}
	
	/* 
	   If answers for all probes have been received,
	   we are finished and must not wait for the timeout.
	*/
	
	if(context->nprobes_received == context->nprobes_sent)	
		return FINISHED;
	else
		return CALL_AGAIN;

}


static void udp_context_fini(struct trigger_instance *this)
{
	struct udp_context *udp_context = this->context;

	kfree(udp_context->ports);
	kfree(udp_context);
}


static int udp_is_response(struct sniffed_packet_descr_t *descr,
			   struct packet_batch *batch)
{
	struct trigger_instance *method_instance = 
		batch->trigger_node;
	
	struct udp_context *udp_context = 
		method_instance->context;
	int port_expected, port_received;
	boolean is_response;

	if( (!udp_context->ports) || (udp_context->nports != 1)){
		printk("INVALID method_instance\n");
		return FALSE;
	}
	
	
	port_expected = udp_context->ports[0];
	port_received = descr->port;


	is_response =  (descr->protocol == IPPROTO_ICMP) &&
		(descr->icmp_type == ICMP_DEST_UNREACH) &&
		(descr->icmp_code == ICMP_PORT_UNREACH) &&
		(port_expected == port_received) &&
		(descr->udp_data == batch->port_indices[0]);
		
	return is_response;

}

static int udp_extract_batch_id(struct sniffed_packet_descr_t *descr)
{
	if ((descr->protocol == IPPROTO_ICMP) &&
	    (descr->icmp_type == ICMP_DEST_UNREACH) &&
	    (descr->icmp_code == ICMP_PORT_UNREACH))
		return descr->udp_data;
	
	return -1;
}



static void udp_register_batch_id(struct trigger_instance *this,
				  unsigned int batch_id)
{
	struct udp_context *context = this->context;
	
	/* Save batch_id in id */
	context->data = batch_id;	

}

static void *
udp_context_copy(struct trigger_instance *this)
{
	struct udp_context *context = this->context;
	
	struct udp_context *new_context = 
		kmalloc(sizeof(struct udp_context), GFP_KERNEL);

	if(!new_context)
		return NULL;

	memcpy(new_context, context, sizeof(struct udp_context));
	
	
	new_context->ports = kmalloc(sizeof(int) * context->nports, GFP_KERNEL);

	if(!new_context->ports){
		kfree(new_context);
		return NULL;
	}

	memcpy(new_context->ports, context->ports, sizeof(int) * context->nports);	
	
	return new_context;
	
}

static int udp_get_round_by_descr(struct sniffed_packet_descr_t *descr)
{
	return descr->port;
}

static int udp_get_round(void *context)
{
	struct udp_context *c = context;

	return c->ports[0];
}

struct trigger_finding_method udp_method = {
	.name = "UDP",
	.sender_func = &udp_sender,
	.receiver_func = &udp_receiver,

	.is_response = &udp_is_response,

	.register_batch_id = &udp_register_batch_id,
	.extract_batch_id  = &udp_extract_batch_id,
	
	.get_round_by_descr = &udp_get_round_by_descr,
	.get_round          = &udp_get_round,

	.context_init = udp_context_init,
	.context_fini = udp_context_fini,

	.context_copy = udp_context_copy,
	
};
